; Wang 2200 MXD Terminal Mux card source code.
; Disassembled and commented by by Jim Battle. All errors are mine.
; Last update: 2018/10/21
;
; TODO:
;    Some of the comments were written while much of the code was a mystery.
;    Everything needs a scrubbing in light of subsequent discoveries.
;
; --------------------------------------------------------------------------
; 2536DWTerminalAndTerminalControllerProtocol.pdf describes the communication
; protocol used between the MXD controller and the terminal.
;
; Controller to terminal protocol:
;    <FB> is the escape character
;    <FB><count><char>  -- repeated character       (count 0x01 to 0x5F)
;    <FB><count + 0x60> -- repeated space character (count 0x01 to 0x5F)
;    <FB><count + 0xCx> -- x/6 second delay         (x is 0x0 to 0x9)
;    <FB><D0>           -- literal <FB> character   (can't be sent compressed)
;    <FB><F0>*          -- route subsequent data to crt
;    <FB><F1>*          -- route subsequent data to printer
;    <FB><F2>*          -- restart terminal
;    <FB><F6>*          -- reset crt only
;    <FB><F8>           -- stop  cursor blinking
;    <FB><FC>           -- start cursor blinking
;
; The '*'-marked commands are "immediate" meaning they can arrive any time,
; even in the middle of a non-immediate sequence, eg
;     <FB><count>(sequence interrupted here)
;                <FB><F1><printer data...><FB><F0>
;                                                 <compressed char>
;
; Terminal to Controller protocol:
;    <12>      -- SHIFT RESET was pressed
;    <13>      -- halt/step key pressed
;    <81>      -- CLEAR key
;    <82>      -- RUN key
;    <84>      -- CONTINUE key
;    <A0>      -- PRINT key
;    <A1>      -- LOAD key
;    <BD>      -- edit key
;    <BE>      -- lowercase FN
;    <BF>      -- uppercase FN
;    <C0>-<DF> -- special function keys
;    <E5>      -- line erase
;    <F8>      -- crt go       (crt buffer ready to accept more data)
;    <F9>      -- printer go   (print buffer ready to accept more data)
;    <FA>      -- crt stop     (terminal not ready to accept more crt data)
;    <FB>      -- printer stop (terminal not ready to accept more printer data)
;
;    The following byte pairs are atomic, meaning no other character can appear
;    between the two bytes:
;    <FC><char> -- transparent: interpret character literally
;    <FD><char> -- EndI/Atom: if char is less than <80>, it is s.f. key
;                               not sure why since we have <C0>-<DF> above
;                             if char is more than <80>, it is an atom
;    <FE><char> -- dead key: like a transparent char, but with special semantics
;
;    page 4 of the pdf says this about "CRT GO":
;       The MXD uses the first occurrence of "CRT GO" to distinguish a 2236D
;       terminal from a non-Wang terminal. The MXD will only send compressed
;       data to a Wang terminal. [ Ed.: see WANGTERM variable ]
;
;    page 4, section 2.7.4, says:
;      When the MXD is powered on, it assumes that the terminal's print buffer
;      is full and the CRT buffer is empty. ...
; --------------------------------------------------------------------------

; The 8080 frequency is 15.9774/9 MHz, or about 1.78 MHz.
;
; The UARTs are i8251.
;
; MEMORY MAP
; 0000 - 0FFF : 4K : ROM
; 2000 - 2FFF : 4K : RAM
;
; Each port gets a block of RAM, minus some holes for stack and globals.
;     uart 1: 0x2000-23FF
;     uart 2: 0x2400-27FF
;     uart 3: 0x2800-2BFF
;     uart 4: 0x2C00-2FFF
;
; p5 of 2236MXE_Documentation.8-83.pdf says this is the MXD buffer usage
;     KEYBOARD buffer:  36 bytes
;     LINE_REQ buffer: 480 bytes
;     PRINTER buffer:  160 bytes
;     CRT buffer:      250 bytes
;
; Here is my top-level reconstruction of how RAM is used. "control" is
; the state associated with managing buffer. Each 1KB block also has
; a hole in it where globals/stack live.
;
;     region            control    buffer     unused?    ctl + buff + unused
;     ----------------  ---------  ---------  ---------  -------------------
;                       2000-2002                         3
;     UART tx buffer    2003-2004  2005-2015              2 +  17
;     KEYBOARD buffer   2016-201A  201B-205A              5 +  64
;     LINE_REQ buffer   205B-206A  206B-2252             16 + 488
;     global state      2254-2265             2253       18 +       1
;     PRINTER buffer    22FE-22FF  2266-22F9  22FA-22FD   2 + 148 + 4
;     CRT buffer        23FE-23FF  2300-23F9  23FA-23FD   2 + 250 + 4
;                                                        ------------
;                                                        1024 bytes

;------------------------------------------------------------------------------
; rough overview of how the MXD operates
;
; The MXD supports up four terminals. The board decodes ports 01-07 (or 41-47,
; 81-87, or C1-C7). An onboard 8080 polls the host cpu interface for CBS/OBS
; strobes as well as requests for data. It manipulates the ready/busy (RBI)
; pin to control the handshake with the host CPU. At that level, the board
; is a "soft peripheral" in that it is 8080 software that determines how the
; CPU sees it.
;
; The 8080 program is 4KB of code in EPROM; the board has 4KB of RAM, which
; is divided up as one 1KB block for each terminal. Each block has a few
; regions not associated with the terminal which hold global state.
;
; BUFFERS
; There are five buffers associated with each terminal.
;    + RX buffer (32 entries of two bytes each, circular)
;      This is a circular buffer of keys received from the terminal.
;      Each entry is two bytes because more than 256 types of events
;      may need to be saved.
;    + CRT buffer (250 bytes, not circular)
;      The host sends data to the crt via OBS strobes to port 05 (VP mode)
;      or port 07 (MVP mode). The MXD itself may want to send bytes to
;      the CRT as well, for example, echoing keys from the terminal.
;      These get queued, then the MXD stops accepting any more data on
;      port 05/07 until all the bytes have been transferred.
;    + PRT buffer (150 bytes, not circular)
;      Much like the CRT buffer, the host cpu sends bytes via OBS to port 04.
;    + UART tx buffer (16 bytes, not circular)
;      The MXD must send both crt and prt buffer traffic over the same
;      serial line. The MXD will transfer up to 16 bytes at a time from
;      either the crt or prt buffer to the tx buffer when the tx buffer
;      goes empty. A polling loop in the MXD will transfer a byte from
;      a tx buffer to the corresponding UART once the UART says it is ready
;      for more data.
;    + line request buffer (488 bytes)
;      This buffer holds the data being input and possibly edited by the
;      user. The MXD handles echoing the characters, buffer editing, and
;      CRT refresh to effect the edits to keep the display in sync with
;      the contents of the buffer.
;
; INTERRUPTS
; Any time one of the four UARTs has received a byte from its terminal, it
; raises an interrupt. The interrupt routine on the 8080 will fetch the
; just received byte, do a bit of preprocessing on it, then push it into
; the RX buffer. Speed is of the essence.
;
; MAIN EVENT LOOP
; MAIN_EVENT is the top level routine which executes forever, polling for
; activity from the host CPU and shuttling data from the output buffers
; to the UART output port. The outline of the main_event loop is this:
; while (true) {
;     - reset stack
;     - call CHECK_HOST routine: poll for OBS or CBS from host
;     - main2: check which uart tx buffers are empty;
;              call TX_POLL to fill them from the crt/prt buffers
;     - txd_lookup: foreach terminal, transfer a byte from the uart tx buffer
;                   to the corresponding UART if there is room
;     - call CHECK_HOST routine: poll for OBS or CBS from host
;     - if there is any line_req in progress and the corresponding rx buffer
;       has data, drop into the RX_POLL routine
; }
;
; TX_POLL
; foreach (terminal with an empty uart tx buffer) {
;     Move data from either the crt or prt buffer to the uart tx buffer.
;     Routing control bytes are inserted in the stream to multiplex between
;     crt and prt data.
; }
;
; RX_POLL
; foreach (terminal with an active line_req and pending key data) {
;     - if there is a "continuation" routine pending jump to it
;     - else dispatch the key according to value. the routine which handles
;       the key may push bytes to the crt buffer to update the display
; }
; jump back to MAIN_EVENT
;
; CHECK_HOST
; This routine is called twice in each iteration of the MAIN_EVENT loop.
; It polls the I/O status registers to determine if the board is addressed
; and if a CBS or OBS strobe happened, or or if the host CPU is wait for
; a response (!CBP). For example, an OBS to port 07 pushes the byte into
; the terminal's CRT buffer, while an OBS to port 04 pushes the byte into
; the PRT buffer.
;
; Where things get interesting is a CBS strobe to port 06 initiates a command
; sequence. The CBS is typically followed by one or more OBS strobes which
; supply parameters to the command. These sequences from the host are delivered
; as quickly as the bus can support them (and the 8080 can handshake them),
; so the handler simply directly polls for the OBS bytes; there is no need
; to suspend in the middle to service TX_POLL or RX_POLL work (though interrupts
; are still active). Once the command sequence is received, the 8080 may need
; to do some work (eg, reset a given terminal, clear a buffer). But the most
; interesting commands are those associated with LINE_REQ, which is the main
; reason why MXD exists.
;
; When the bit corresponding to the current terminal has its bit set in the
; LREQ_ACCEPT byte and a key is waiting to be processed, the key gets interpreted
; in light of the active command. Most keystrokes just represent ordinary data
; input, so the byte is appended into the line_req buffer and the character is
; pushed into the crt buffer fifo to effect the echo. But some keys have a
; special meaning and require more work. For instance in EDIT mode, the S.F. 10
; key means "insert a character". The 8080 must shift all the bytes from the
; current cursor position to the end of the line_req buffer right one position;
; bytes must be sent to the crt to redraw the affected portion of the buffer
; (keeping in mind line wrapping and such); then the cursor must be repositioned
; to the active cursor position.
;
; Some redrawing commands make take a long time or require stuffing many bytes
; into the crt buffer. As a result, these routines have to be broken up into
; chunks of work. The byte at 0x2n5C is normally 00, but if it is not, the
; RX_POLL will take a detour to a "continuation" of the suspended work to allow
; more progress. One command may have many continuations chained together
; (see the first half of the INDTAB to see them).

;------------------------------------------------------------------------------
; What follows is a map of how memory is allocated for the first terminal.
; (add 0x0400*n for terminal N-1)
;
; 0x2000    terminal associated with this block, one-hot (01=term 1, 02=term 2, 04=term 3, 08=term 4)
; 0x2001    ? per-term RBI status? (see code after REQ_PRINT_BUFFER, and SET_ACT_TERM)
; 0x2002    terminal status byte (see STATUS_RESPONSE routine)

; There is a buffer for data destined for the crt at 0x2300-0x23F9, and a buffer
; for printer data at 0x2266-0x22F9. the tx channel needs to multiplex both of
; these streams onto the one serial port. It does so by prefixing crt data with
; the <FB><F0> escape, and printer data by <FB><F1>. It would be inefficient to
; switch too often, so the mxd batches up work of 8-16 (ish) bytes at least.
; this data is saved in the uart tx buffer at 0x2005-0x201F. This is not a
; circular buffer. [2n03] is the count of bytes remaining to send, and [2n04] is
; the get pointer (just one byte, since it is all in the current page). Once
; all the data is transmitted, [2n03] is reinitialized to 0x00 and [2n04] to 0x04.
; the latter is surprising, it is pre-incremented before use, so get=0x04 really
; means the next byte to send is at 0x2n05.
;
; 0x2003         remaining bytes to send in uart tx buffer
; 0x2004         uart tx buffer get pointer (next byte to send is at get+1)
; 0x2005-0x201F  uart tx buffer (27 bytes, though I'm not sure how full it gets)
;                I suspect the crt is allowed to save 14 bytes, and the printer 8,
;                plus two bytes each to establish mode (2+14+2+8) = 26

; pending keystroke circular buffer.
; as characters are received from the terminal, they are saved in this buffer.
; they aren't processed until the 2200 issues some type of command which may
; consume them (eg, LINE_REQ, or KEYIN_LINE_REQ).
; The first five bytes (2n16-2n1A) are zero-filled in INIT_CRT_BUF.
;     0x2016 - saved keycode value
;     0x2017 - saved prefix byte, waiting for 2nd byte to arrive
;     0x2018 - count of entries in the rx buffer
;     0x2019 - put pointer is left pointing to the 2nd byte of the most recent
;              entry, which means it pre-increments on push
;     0x201A - get pointer to most recent entry of the rx buffer
; 0x201B..5A - is a circular buffer of 32 two-byte entries. the first (prefix)
;              (prefix) byte indicates the class of the second byte. the second
;              byte is the key value. the prefix byte is mostly the same as
;              documented on page 4 of this pdf, but with a FF type added:
;                  2536DWTerminalAndTerminalControllerProtocol.pdf
;           0xFC: transparent -- interpret keycode literally, no atom or escape processing
;           0xFD: EndI/Atom   -- <= 0x7F is special function key; >= 0x80 is an atom
;           0xFE: dead key    -- char echoed, then the cursor is backed up one,
;                                but otherwise identical to transparent key handling
;           0xFF: normal      -- this isn't sent on the line, but is placed into the kb buff;
;                                it tags the value as being none of the above cases, I guess

; This section is associated with procesing LINE_REQ and any of the associated commands.
; 0x205B         unused?
; 0x205C         "continuation offset" (see the comment in SAV_CONT)
; 0x205D/5E      big endian pointer to next byte of line_req buffer that needs to be redrawn
; 0x205F/60      big endian pointer to last byte of line_req buffer that needs to be redrawn
; 0x2061         key prefix byte; used in SF15 (recall) handling, and in QUERY_LINE_REQ
; 0x2062         key code value;  used in SF15 (recall) handling, and in QUERY_LINE_REQ
; 0x2063         line request flag byte
;                    01: flush previously stored keystrokes
;                    02: keyin_line_req command flag
;                    04: EDIT mode
;                    08:
;                    10: refill_line_req command flag
;                    20: LINPUT mode   (learned from DAMVP35)
;                    40: OS > 1.3      (learned from DAMVP35)
;                    80: underline
; 0x2064         line_req cursor column
; 0x2065/66      line_req pointer (big endian) to end of buffer
; 0x2067/68      pointer (big endian) to last used byte of line_req buffer,
;                perhaps one past the last byte used. I haven't checked carefully.
; 0x2069/6A      pointer (big endian) to current cursor location in line_req buffer
; 0x206B-0x2252  line request buffer (0x1E8=488 characters max)

; 0x2266-0x22F9  prt buffer; data waiting to be sent to the terminal printer (148 bytes)
; 0x22FE         print buffer put pointer; set to 0x76 in INIT_PRT_BUF routine (note: offset byte after area zero'd out above)
; 0x22FF         print buffer get pointer; set to 0x66 in INIT_PRT_BUF routine (note: offset byte of area zero'd out above)
;
; 0x2300-0x23F9  crt buffer; data waiting to be sent to the terminal crt (250 bytes)
; 0x23FE         crt buffer put pointer
; 0x23FF         crt buffer get pointer

;------------------------------------------------------------------------------
; global locations, not per-terminal

START_OF_RAM	equ	02000H

; 0x2252-2265 should be free (fits between end of line_req buffer and start of prt buffer)

; 0x2266-0x22F9  prt buffer; data waiting to be sent to the terminal printer (148 bytes)
MVP_MODE	equ	02256H	; 0x00=vp mode, 0xFF=mvp mode
SHADOW_RBI	equ	02257H	; most recent value stored in RBI register

; CUR_BLOCK is a 16b pointer to the base of the 1KB block of memory associated
; with the terminal most recently selected by SET_ACT_TERM. It takes one of
; these values:  0x2000, 0x2400, 0x2800, or 0x2C00.
; CUR_BLOCK_PAGE is just the ms byte, namely 0x20, 0x24, 0x28, or 0x2C.
CUR_BLOCK	equ	02258H
CUR_BLOCK_PAGE	equ	02259H

; CUR_TERM is closely related to CUR_BLOCK variable; it is a one-hot encoding
; of the terminal most recently selected by SET_ACT_TERM. It takes one of
; these values: 0x01, 0x02, 0x04, 0x08.
CUR_TERM	equ	0225AH

CUR_UART	equ	0225CH	; one-hot bitmap to select default uart
HAS_UART_TXD	equ	0225EH	; bitmap of which ports have data in their uart tx buffer

; a map which ports need to be serviced (1=port 1, 2=port 2, 4=port 3, 8=port 4)
; in TX_POLL, it is which terminal is having crt or prt data transferred to the uart tx buffer.
; in RX_POLL, it is which terminal is processing keyboard buffer entries.
PENDING_TERMS	equ	02656H	; terminal bitmap

; LINE_REQ_BLOCK is akin to CUR_BLOCK in that it holds a pointer to the 2n00
; base address for a 1K block of RAM. CUR_BLOCK always points to the base of
; the terminal current addressed. LINE_REQ_BLOCK does too during the QUERY_LINE_REQ
; and ACCEPT_LINE_REQ_DATA commands, but during RX_POLL processing it points at
; whichever terminal is being serviced in that iteration of the RX_POLL loop.
LINE_REQ_BLOCK	equ	02657H	; memory block for most recent line request. how is this different from CUR_BLOCK?
LINE_REQ_PAGE	equ	02658H	; ms byte of LINE_REQ_BLOCK pointer

ULINE_FLAG	equ	02659H	; underline flag (either 0x00 or ox80)
BUF_ACTIVE_IND	equ	0265AH	; pointer to either CRT_BUF_ACTIVE or PRT_BUF_ACTIVE
CUR_CAN_CMPRS	equ	0265CH	; non-zero means current terminal supports Wang compression
TX_SERVICE	equ	0265DH	; term being serviced by tx polling. [265D] = crt/printer map; [265E] = page0 for term

; note that two different bits of code use the same 265F/60 pair for their own purposes
; it stores a pointer to data associated with terminal currently being serviced.
CUR_INIT 	equ	0265FH	; 2n00 (in INIT_CRT_BUF/INIT_PRT_BUF)
CUR_TX_BUF	equ	0265FH	; 2n03 (in TX_POLL)

TX_RECEIVER	equ	02661H	; bitmap: current routing target of tx bytes: 0=crt, 1=printer

; The data for the crt and printer associated with a given terminal are
; multiplexed on the same serial line. These two bytes act heuristically to
; reduce the effect of the printer traffic on the flow of crt traffic. The logic
; is arcane, so this is my best reconstruction of what is going on. Both
; PRT_PRIO and SENT_CR are bitmaps, one bit per terminal in the same fashion as
; other bitmaps here. in the description below when I say something is set or
; cleared, I just mean the bit associated with the current terminal.
;
; SENT_CR is reset to zero. If, during the middle of a transfer of bytes from
; the crt or prt buffer to the uart tx buffer, a carriage return is found
; (see CR_LIMIT_XFER:), this bit gets set. Once the remainder of the crt or
; prt buffer is transferred to the tx buffer (see txp4:), this bit is cleared
; back to zero. Its state is used only to affect PRT_PRIO.
;
; A confusing thing is that the logic which twiddles these bytes is largely
; independent of whether we are transferring crt or printer data. Eg, when a
; <CR> is sent to the uart tx buffer, it affects crt/prt transmission priority,
; but the logic doesn't care if it was a <CR> for the printer or the crt.
;
; PRT_PRIO=1 is used by the logic in main2 to prioritize prt traffic over the
; serial line if both crt/prt buffers have data to send.

; PRT_PRIO is
;   - reset to zero.
;   - in main2:, if the uart tx buffer is empty and the prt buffer is ready
;     to send (non-empty and not flow-controlled), this bit gets cleared.
;   - if, during the middle of a transfer of bytes from the crt or prt buffer
;     to the uart tx buffer, a carriage return is found (see CR_LIMIT_XFER:),
;     this bit gets set (as does SENT_CR).
;   - in txp4:, the SENT_CR state comes into play; when the crt or prt buffer
;     is completely drained to the tx buffer, this bit is set if the prt buffer
;     is ready to send data and the prt/crt buffer hasn't sent a CR since the
;     last time the prt/crt buffer was empty
; Perhaps the reason for the last bullet above is to promote a prt transfer
; every other tx chunk sent.

PRT_PRIO	equ	02662H	; terminal bitmap; see the description above
SENT_CR		equ	02663H	; terminal bitmap; see the description above

; these four bytes are analogous to same for crt
PRT_STAT_PEND	equ	02A56H	; which terms were queried for REQUEST_CRT_BUFFER but were not empty at the time
PRT_CAN_TX	equ	02A57H	; which terms are ready to tx (1=non-empty && not flow controlled)
PRT_BUF_ACTIVE	equ	02A58H	; which terms have non-empty prt buffer (1=non-empty)
PRT_STOPPED	equ	02A59H	; which terms are in prt flow control state (1=stopped)

; the following 10 bytes each contain a terminal bitmap of some status,
; where 01=term 1, 02=term 2, 04=term 3, 08=term 4

WANGTERM	equ	02A5AH	; 1=flow control seen, so term is wang-compatible

; cleared on POR or INIT_CUR_TERM or reset key
; set on reception of STOP flow control (independent of CRT or PRT type)
; this is probably the "Wang Terminal" qualifier
;       The MXD uses the first occurrence of "CRT GO" to distinguish a 2236D
;       terminal from a non-Wang terminal. The MXD will only send compressed
;       data to a Wang terminal.

; these four bytes are analogous to same for printer
CRT_STAT_PEND	equ	02A5BH	; which terms were queried for REQUEST_CRT_BUFFER but were not empty at the time
CRT_CAN_TX	equ	02A5CH	; which terms are ready to tx (1=non-empty && not flow controlled)
CRT_BUF_ACTIVE	equ	02A5DH	; which terms have non-empty crt buffer (1=not empty)
CRT_STOPPED	equ	02A5EH	; which terms are in crt flow control state (1=stopped)

KB_BUF_READY	equ	02A5FH	; which terms have non-empty keyboard buffer (1=not empty) or a continuation pending


; These flags encode important state about line_req processing for each terminal.
; The comment block here is extensive because it took a lot of reverse engineering
; to figure out what they mean (assuming they are right at all).
;
; LREQ_ACTIVE is true from when the END_LINE_REQ command is issued
;    (n.b.: don't be mislead by the name; it means the setup for the line
;     request is done, and the MXD should start processing the command) until
;    the host abandons the command via DELETE_LINE_REQ or the term is reset.
;    It is reset to 0
;          set     in END_LINE_REQ
;          cleared in DELETE_LINE_REQ
;          tested  in QUERY_LINE_REQ   to determine response code
;
; LREQ_ACCEPT should never be true unless LREQ_ACTIVE is also true. When 0 it
;    indicates that processing is done, either because of DELETE_LINE_REQ, or
;    the user enters a key which terminates processing (CR, s.f. key, or recall).
;    When 1 it indicates either that the RX_POLL loop should perform a
;    "continuation" command or it should dispatch any key it receives.
;    It is reset to 0
;       is set     in END_LINE_REQ
;       is set     in alrd6:           to initiate continuation callbacks
;       is cleared in DELETE_LINE_REQ
;       is cleared in L0D43  << called when RX_POLL processes 0D, or <FF><nn> (nn>E6), or <FD><s.f.>
;       is tested  in main5; if ==0 then MAIN_EVENT, else perform RX_POLL
;       is tested  in QUERY_LINE_REQ
LREQ_ACCEPT	equ	02A60H	; which terms have line_req still in progress (active, but not terminated; see QUERY_LINE_REQ)
LREQ_ACTIVE	equ	02A61H	; which terms are doing LINE_REQUEST (1=active)

RESET_MAP	equ	02A62H	; which terms have hit RESET (1=reset)
HALT_MAP	equ	02A63H	; which terms have hit HALT  (1=halt)

; The stack is on INTERRUPT, 10 bytes are saved on the stack. Non-interrupt code
; can get as deep as four call/push layers:
;     MAIN_EVENT -> call CHECK_HOST -> call COMPRESS -> call TIMED_POLL -> push h
; So 8 bytes for normal code, and 10 for interrupt = 18 bytes.
; Actually, I don't see that 2E53 is used, but one extra byte for the stack doesn't help.
TOP_OF_STACK	equ	02E66H	; stack occupies 2E65 -> 2E54

INIT_TYPE	equ	02FFDH	; either 0x03 (clear screen) or 0x07 (bell)
				; sent after initializing a given terminal

; input port defines
IN_UART_TXRDY		equ	000H	; parallel poll of which UARTs have room in their tx fifo
IN_2200_STATUS		equ	001H	; various status bits
	;   0x01 = OBS strobe seen
	;   0x02 = CBS strobe seen
	;   0x04 = PRIME (reset) strobe seen (cleared by OUT 0)
	;   0x08 = high means we are selected and the CPU is waiting for input
	;   0x10 = board selected at time of ABS
	;   0x20 = AB1 when ABS strobed
	;   0x40 = AB2 when ABS strobed
	;   0x80 = AB3 when ABS strobed
IN_OBUS_N		equ	002H	; read !OB bus, and clear obs/cbs strobe status
IN_OBSCBS_ADDR		equ	003H	; [7:5] = [AB3:AB1] at time of cbs/obs strobe (active low?)
IN_UART_RXRDY		equ	004H	; parallel poll of which UARTs have received a byte
IN_UART_DATA		equ	006H
IN_UART_STATUS		equ	00EH
	;   0x80 = DSR       (data set ready)
	;   0x40 = BRKDET    (break detect)
	;   0x20 = FE        (framing error)
	;   0x10 = OE        (overrun error)
	;   0x08 = PE        (parity error)
	;   0x04 = TxEMPTY   (the tx fifo is empty and the serializer is done)
	;   0x02 = RxRDY     (a byte has been received)
	;   0x01 = TxRDY     (the tx fifo buffer has room for a character)

; output port defines
OUT_CLR_PRIME	equ	000H	; clears reset latch
OUT_IB_N	equ	001H	; drive !IB1-!IB8, pulse IBS
OUT_IB9_N	equ	011H	; same as OUT_IB_N, plus drive IB9
OUT_PRIME	equ	002H	; fires the !PRIME strobe
OUT_HALT_STEP	equ	003H	; one shot strobe
OUT_UART_SEL	equ	005H	; uart chip select
OUT_UART_DATA	equ	006H	; write to selected uart data reg
OUT_RBI		equ	007H	; 0=ready/1=busy; bit n=addr (n+1); bit 7=n/c
				; 01=1(kb),02=2(status),04=3(n/c),08=4(prt),10=5(vpcrt),20=6(cmd),40=7(crt)
OUT_UART_CMD	equ	00EH	; write to selected uart command reg

; The RBI register is a bitmap indicating whether a given address offset is
; ready or not. Unfortunately, addr n isn't just bit n.
RBI_ADDR_1	equ	001H
RBI_ADDR_2	equ	002H
RBI_ADDR_3	equ	004H
RBI_ADDR_4	equ	008H
RBI_ADDR_5	equ	010H
RBI_ADDR_6	equ	020H
RBI_ADDR_7	equ	040H

;------------------------------------------------------------------------------
; In the code below, Lxxxx are autogenerated labels from the original disassembly.
; Labels use this prefix:
;    Lxxxx = address xxxx, target of multiple jumps/calls
;    sxxxx = single entry point
;    txxxx = local target (no far jumps or calls)
; TBD: ideally, all targets would have a semantic label.
;
; Labels using upper case are call/jump targets from distant code.
; Lowercase labels signify local branch targets from related code.
; This helps better understand the program structure.

	org	00000H
coldstart:			; nothing branches here; reset only
	di
RESTART:
	mvi	a,0FFH
	out	OUT_RBI		; all addresses are busy
	jmp	COLD_INIT

; save A to SHADOW_RBI, then send it to !RBI register
RST1:				; depth=1
	sta	SHADOW_RBI
	out	OUT_RBI
	ret

; mystery -- perhaps it encodes the ROM version number
HLF001:
	db	001H
	db	03EH

; like RST1, but enable interrupts before returning
; save A to SHADOW_RBI, send it to !RBI register, enable interrupts
RST2:				; depth=1
	sta	SHADOW_RBI
	out	OUT_RBI
	ei
	ret
	nop

; check if OBS strobe has arrived.
; if it has, read the data bus, convert to proper polarity, and return.
; if it hasn't and the address is port 6, keep polling for OBS.
; if the address is other than 6, jump to the MAIN_EVENT loop to keep
; the movement of data to the terminals.
RST3:				; depth=1
	in	IN_2200_STATUS
	rrc
	jnc	NO_OBS		; branch if no OBS strobe
	; fall through ...

; fetch the data that came with OBS/CBS, clear strobes
GET_OBS_DAT:			; depth=1
	in	IN_OBUS_N	; get !OB[7:0], clear OBS/CBS latches
	cma			; true polarity
	ret

	nop
	nop
	nop
	nop

; emit bell tone to CRT output buffer
BEEP:				; depth=2
	mvi	a,007H
;       fall through ...

; Save A in the crt output buffer
; This buffer isn't circular; no wrapping is done. It is the responsibility
; of the caller to check for space before calling.
RST5:				; depth=2 (call here, push h)
	push	h
RST5b:				; depth=1
	lhld	LINE_REQ_BLOCK
	inr	h		; next page
	inr	h		; next page
	inr	h		; page 3
	mvi	l,0FEH		; 23FE+n*0x0400: CRT buffer put pointer
	inr	m		; bump it to next byte
	mov	l,m
	dcr	l
	mov	m,a		; save A at the ptr location before the bump
				; that is the put pointer is where the next one will go
	; fall through ...

; pop hl and return (used only by the fall-through and A_TO_LINE_REQ)
POPRET:
	pop	h
	ret
	nop

; An interrupt is raised any time the RxRdy line of any of the four UARTs go
; high, indicating a character has been received. Most of the work is done by
; INT_HANDLER.
INTERRUPT:
	di
	push	h
	push	d
	push	b
	push	psw
	lxi	h,0
	dad	sp
	mvi	a,02EH		; sanity check. if SP is under 2E00, something
	cmp	h		; ... is terribly wrong. The stack occupies ...
	jz	INT_HANDLER	; ... only 0x2E65->0x2E56 (16 bytes).
	jmp	RESTART		; abort everything and reset

; OBS latch isn't set
NO_OBS:
	rlc		; get original IN_2200_STATUS back
	ani	0FBH	; clear the reset status bit (PRIME)
	cpi	0D0H	; 1101 = check for ab[3:1]=6, board selected; 0000 = no cbs, no obs
	jz	RST3	; keep polling if addr 6
	jmp	MAIN_EVENT

;------------------------------------------------------------------------------
; zero fill (hl) for de bytes, kind of (starting hl must be even aligned)
; really it fills from (hl) to the end of the current page, and each subsequent
; page for D times. Then the first E bytes of the next page are zero filled.
; But there is only one caller, and the initial HL is page aligned.
ZFILL_HL_DE:		; depth=1
	xra	a
zfill2:
	mov	m,a
	inr	l
	mov	m,a
	inr	l
	jnz	zfill2
	inr	h	; another page ...
	dcr	d	; ... has been filled
	jnz	zfill2
	mov	b,e
	; ... fall through and do the remaining partial page

; zero fill (hl) with B bytes of memory
ZFILL_HL_B:		; depth=1
	xra	a
	; fall through ...

; fill (hl) with B bytes of memory with the value in A
A_FILL_HL_B:		; depth=1
	mov	m,a
	dcr	b
	inx	h
	jnz	A_FILL_HL_B
	ret

;------------------------------------------------------------------------------
; fetch the 16b pointer in memory pointed at by (256*B + L) into HL
; compare that to DE. return with larger in HL and smaller in DE.
; (in practice, the caller always sets L to 0x67)
HLPTR_MAXMIN_DE:	; depth=1
	mov	h,b

; fetch the 16b pointer in memory pointed at by HL into HL.
; compare that to DE.
; return with larger in HL and smaller in DE.
HLPTR_MAXMIN_DE2:
	mov	a,m	; HL = pointer stored at HL
	inx	h
	mov	l,m
	mov	h,a
	; fall through ...

; compare HL and DE
; return with larger in HL, smaller in DE
; a better way (7 bytes instead of 10) would have been:
;     mov a,e
;     sub l
;     mov a,d
;     sbc h
;     rnc      return if hl >= de
;     xchg
;     ret
MAX_HL_MIN_DE:
	mov	a,h
	sub	d
	jnz	mhlmde2
	ora	l
	cmp	e
mhlmde2:
	rnc
	xchg
	ret

; save A to the current CRT output buffer.
; it is different than a naked "rst 5" in that it remaps non-printable chars,
; and sets the underline bit if ULINE_FLAG is 0x80.
A_TO_CRT_BUF:			; depth=2 (call here, push h)
	push	h
	mov	h,a
	lda	ULINE_FLAG
	ora	h
	cpi	010H		; is it a control character?
	jnc	atcb1		; no
	mvi	a,02EH		; yes, use '.' instead
atcb1:
	cpi	0FBH		; is it a compression escape char?
	jnz	RST5b		; do RST 5 but don't return
	rst	5		; put FB in the terminal output buffer
	mvi	a,0D0H
	jmp	RST5b		; FB D0 sequence means literal "FB" byte

; set terminal response byte to indicate keyboard buffer not empty or line request complete
; (see STATUS_RESPONSE)
SET_KB_RDY_STATUS:
	mvi	a,004H		; keyboard buffer not empty or line request complete (see STATUS_RESPONSE)
	; fall through ...

; set terminal response byte according to reg A (see STATUS_RESPONSE)
SET_STATUS_RESP:		; depth=1
	mvi	l,002H		; 2n02 = terminal status response byte
	ora	m
	mov	m,a
	; fall through ...

; indicate that the current terminal is ready to send a status response
; note this also disabled interrupts
ADDR2_READY:
	di
	lda	SHADOW_RBI
	ani	NOT(RBI_ADDR_2)	; addr 2 is ready
	rst	1		; save A to SHADOW_RBI, then send it to !RBI register
	ret

; wait for the CPU to poll for input.
; however, if the current addr is not 1, 2, or 6, reset the terminal state
; we have built as we are confused and can't trust it.
CPU_POLL_WAIT:			; depth=1
	in	IN_2200_STATUS
	ani	008H		; test if CPU is waiting for input from us
	rnz			; return if it is
	in	IN_2200_STATUS
	ani	0F3H		; ignore PRIME and !CPB
	cpi	050H		; check addr=2, board selected, !CBS, !OBS
	jz	CPU_POLL_WAIT
	cpi	0D0H		; check addr=6, board selected, !CBS, !OBS
	jz	CPU_POLL_WAIT
	cpi	030H		; check addr=1, board selected, !CBS, !OBS
	jz	CPU_POLL_WAIT
	jmp	MAIN_EVENT	; the board isn't selected, AB=3/4/5/7, CBS, or OBS

	nop			; alignment padding

;------------------------------------------------------------------------------
; this is the jump table that maps addr 6 CBS command bytes to action;
; it is also used by code at RX_POLL to continue certain commands by branching
; based on the contents of 2n5C (continuation byte).
; see code at IDXJMPA for how this table is addressed.
;
; can't actually put the org here or the assembler chops up the .hex output,
; making comparison to the original more difficult.
;	org	000BCH

INDTAB:
	; this part of the table is used for dispatching command continuations
	; TBD: it isn't clear how the ?? offsets are reached
	dw	CHECK_KB_MT_FLAG	; 00BC - ?? (continuation byte = 05EH)
	dw	CURSOR_ON		; 00BE - continuation#1 of END_OF_LINE_REQ
	dw	BLANK_SPAN		; 00C0 - continuation#2 of END_OF_LINE_REQ
	dw	DO_REDRAW		; 00C2 - continuation#3 of END_OF_LINE_REQ + cont#1 of EDIT_SF9/10
	dw	CURSOR_TO_START		; 00C4 - continuation#4 of END_OF_LINE_REQ + cont#2 of EDIT_SF9/10
	dw	CHECK_KB_MT_FLAG	; 00C6 - continuation#5 of END_OF_LINE_REQ + cont#3 of EDIT_SF9/10
	dw	ERROR_LINE_REQ2		; 00C8 - continuation of ERROR_LINE_REQ
	dw	CURSOR_ON		; 00CA - continuation#1 of TERMINATE_LINE_REQ
	dw	BLANK_SPAN		; 00CC - continuation#2 of TERMINATE_LINE_REQ
	dw	DO_REDRAW		; 00CE - continuation#3 of TERMINATE_LINE_REQ
	dw	CURSOR_TO_START		; 00DO - continuation#4 of TERMINATE_LINE_REQ
	dw	INJECT_CR		; 00D2 - continuation#5 of TERMINATE_LINE_REQ
	dw	CHECK_KB_MT_FLAG	; 00D4 - ?? (continuation byte = 06AH)
	dw	DO_REDRAW		; 00D6 - triggered from alrd6
	dw	SET_LINEREQ_STATUS	; 00D8 - continuation#1 from alrd6

	; this part of the table is used for io6 CBS(nn) dispatch
	dw	CHECK_HOST		; 00DA - addr 6, CBS 12 - ignore (triple controller, mxe only)
	dw	CHECK_HOST		; 00DC - addr 6, CBS 11 - ignore (triple controller, mxe only)
	dw	TERMINATE_LINE_REQ	; 00DE - addr 6, CBS 10
	dw	ERROR_LINE_REQ		; 00E0 - addr 6, CBS 0F
	dw	REQ_PRINT_BUFFER	; 00E2 - addr 6, CBS 0E
	dw	REQ_CRT_BUFFER		; 00E4 - addr 6, CBS 0D
	dw	ACCEPT_LINE_REQ_DATA	; 00E6 - addr 6, CBS 0C
	dw	QUERY_LINE_REQ		; 00E8 - addr 6, CBS 0B
	dw	END_OF_LINE_REQ		; 00EA - addr 6, CBS 0A
	dw	REFILL_LINE_REQ		; 00EC - addr 6, CBS 09
	dw	PREFILL_LINE_REQ	; 00DA - addr 6, CBS 08
	dw	LINE_REQ		; 00F0 - addr 6, CBS 07
	dw	KEYIN_LINE_REQ		; 00F2 - addr 6, CBS 06
	dw	KEYIN_POLL_REQ		; 00F4 - addr 6, CBS 05
	dw	KB_READY_CHECK		; 00F6 - addr 6, CBS 04
	dw	DELETE_LINE_REQ		; 00F8 - addr 6, CBS 03
	dw	INIT_CUR_TERM		; 00FA - addr 6, CBS 02
	dw	POWER_ON_INIT		; 00FC - addr 6, CBS 01
	dw	CHECK_HOST		; 00FE - addr 6, CBS 00 -- null command

;------------------------------------------------------------------------------
	org	00100H		; TERM_LOOKUP requires this be page aligned
TERM_PRIORITY:
	dw	00000H
	dw	02001H
	dw	02402H
	dw	02001H
	dw	02804H
	dw	02001H
	dw	02402H
	dw	02001H
	dw	02C08H
	dw	02C08H
	dw	02402H
	dw	02C08H
	dw	02804H
	dw	02804H
	dw	02402H
	dw	02C08H

; TERM_LOOKUP picks one of four terminals to consider next.
; Eg, it might be a map of uarts which have receive data, or which uarts have
; room in their tx buffer, etc.
; entry:
;     a[3:0] contains a bitmap of terminals to consider
; exit:
;     c[3:0] contains the chosen terminal as a one-hot bitmap.
;     a[3:0] is the same as the original except the chosen bit is cleared.
;     b and h both contain a page pointer to the chosen port block.
;     l is garbage.
;
; Here is the full mapping of all 16 possible inputs:
;     A       C   B     A (exit)
;     0000 -> 00, 00    0000
;     0001 -> 01, 20    0000
;     0010 -> 02, 24    0000
;     0011 -> 01, 20    0010
;     0100 -> 04, 28    0000
;     0101 -> 01, 20    0100
;     0110 -> 02, 24    0100
;     0111 -> 01, 20    0110
;     1000 -> 08, 2C    0000
;     1001 -> 08, 2C    0001
;     1010 -> 02, 24    1000
;     1011 -> 08, 2C    0011
;     1100 -> 04, 28    1000
;     1101 -> 04, 28    1001
;     1110 -> 02, 24    1100
;     1111 -> 08, 2C    0111
;
; The TERM_PRIORITY table is structured like this.
; The first byte priority encodes which terminal to look at.
;    For the first 8 entries, uart 1>2 and 2>3.
;    For the next 8 entries, the pattern isn't evident.
;    Maybe it is just a scrambled code to try to give them all; equal priority
;    over all combinations of active channels. In fact, each channel is given
;    top priority in half of all cases where it's is active.
;
; The second byte = (0x20 + 4*(first_byte-1)), and is a pointer to a page in RAM
; where all the terminal-specific state and buffers live. Typically that value
; is saved in H, and other code fiddles with L to develop an (HL) pointer to
; whatever offset is desired.
;     uart 1: 0x20xx
;     uart 2: 0x24xx
;     uart 3: 0x28xx
;     uart 4: 0x2Cxx
;
TERM_LOOKUP:				; depth=1
	rlc				; double the table index
	mvi	h,high(TERM_PRIORITY)	; page aligned table
	mov	l,a
	mov	c,m			; grab 1st byte of entry
	rrc
	xra	c
	inx	h
	mov	b,m			; grab 2nd byte of entry
	mov	h,b			; h points to pages for this port
	ret

;------------------------------------------------------------------------------

; hl = de-hl, with A destroyed and flags set
DE_SUB_HL:			; depth=1
	xchg
; hl = hl-de, with A destroyed and flags set
HL_SUB_DE:
	mov	a,l
	sub	e
	mov	l,a
	mov	a,h
	sbb	d
	mov	h,a
	ret

; this is called when the board is addressed at offset 5.
; return if we are in VP mode, otherwise reinitialize structs.
SET_VP_MODE:			; depth=1
	lda	MVP_MODE
	ana	a
	rz			; return if we were already in vp mode
	mvi	a,0CFH		; addr 5/6 are ready
	sta	SHADOW_RBI
	xra	a
	sta	MVP_MODE	; indicate VP mode (AKA boot mode)
	mvi	a,007H		; send bell at end of re-init
	jmp	pwron2		; re-init

; only path here is via power on reset.
; "rst 0" would do it too, but that never happens.
COLD_INIT:
	in	IN_OBUS_N	; side effect: clear OBS/CBS latches
	; fall through ...

POWER_ON_INIT:
	mvi	a,003H		; clear crt at end of init
pwron2:
	di
	lxi	sp,TOP_OF_STACK
	sta	INIT_TYPE	; crt initialization char
	lxi	h,START_OF_RAM
	lxi	d,00E64H	; zero fill 0x2000 to 0x2E63 (going any
				; farther would clobber the return stack)
	call	ZFILL_HL_DE

	mvi	c,008H		; process term 4 down to term 1
init_uarts:
	call	INIT_CRT_BUF	; returns with hl=0x2312+n*0x400
	lda	INIT_TYPE
	dcx	h		; 0x2311+n*0x400
	mov	m,a		; either clear screen or bell
	call	INIT_PRT_BUF

	mov	a,c
	sta	CUR_UART
	out	OUT_UART_SEL
	; the i8251 manual, page 2-12, says
	;   When power is first applied, the 8251A may come up in the Mode, Sync,
	;   character or Command format. To guarantee that the device is in the
	;   Command Instruction format before the Reset command is issued, it is
	;   safest to execute the worst-case initialization sequence (sync mode
	;   with two sync characters). Loading three 00Hs consecutive into the
	;   device with C/!D=1 configures sync operation and writes to dummy 00H
	;   sync characters. An Internal Reset command (40H) may then be issued
	;   to return the device to the "Idle" state.
	xra	a
	out	OUT_UART_CMD
	out	OUT_UART_CMD
	out	OUT_UART_CMD
	out	OUT_UART_CMD
	mvi	a,040H		; internal reset command
	out	OUT_UART_CMD
	mvi	a,05EH		; mode byte: 1 stop bit, odd parity, 8 bits, 16x baud rate factor
	out	OUT_UART_CMD
	mvi	a,037H		; command: RTS=1, DTR=1, Rx/Tx en, error reset
	out	OUT_UART_CMD
	in	IN_UART_DATA
	mov	a,c
	rrc
	mov	c,a
	jnc	init_uarts	; process next terminal in sequence

	mvi	a,087H		; addr 4,5,6,7 are ready
	rst	2		; save A to SHADOW_RBI, send it to !RBI register, enable interrupts

	; check sum the eprom for integrity. if it fails, transmit a constant stream of
	; "D" to the terminal, otherwise jump to the main event loop entry point
	mvi	b,010H		; do 16 pages
	lxi	h,0000H		; start of EPROM
	mvi	a,0FFH
	mvi	c,008H
rom_cksum1:
	rlc			; some kind ...
	jnc	rom_cksum2	; ... of ...
	xra	c		; ... lfsr
rom_cksum2:
	xra	m		; checksum the next byte of eprom
	inr	l
	jnz	rom_cksum1	; loop until end of page
	inr	h		; then go to next page
	dcr	b
	jnz	rom_cksum1
	ana	a
	jz	MAIN_EVENT	; jump if checksum is 00 (good)
	dcx	h		; \    my guess is developers use RAM not EPROM,
	xra	m		;  >-- and this fixes the the checksum in RAM
	mov	m,a		; /    for the next restart
	; emit an endless stream of 'D' characters to indicate an MXD EPROM
	; checksum failure
cksum_hang:
	in	IN_UART_STATUS
	rrc			; TxRDY status bit
	jnc	cksum_hang	; jump if no room in the tx fifo
	mvi	a,'D'
	out	OUT_UART_DATA
	jmp	cksum_hang

;------------------------------------------------------------------------------
; Main Event Loop
; give up on any operation that was in progress and reset state for next activity
MAIN_EVENT:
	ei
	lxi	sp,TOP_OF_STACK	; reset stack pointer
	in	IN_2200_STATUS
	cpi	0B0H		; addr=5, selected, !cpb, no reset, no cbs, no obs
	jnz	main2
	in	IN_2200_STATUS	; double check in case we ...
	cpi	0B0H		; ... sampled during transition? (yuk)
	cz	SET_VP_MODE	; if port 5 is addressed, revert to VP mode

main2:
	call	CHECK_HOST	; check if the host cpu needs attention

	; CPU hasn't sent anything and isn't expecting input;
	; check if anything can be sent to the terminals
	lda	HAS_UART_TXD	; bitmap of which ports have data in uart tx buffer
	cma
	mov	b,a		; bitmap, 1=tx buffer is empty

	lxi	h,PRT_CAN_TX
	ana	m		; 1=(printer is ready to send) & (uart tx buffer is empty)

; I think in some way, PRT_PRIO=0 means deprioritize prt behind crt transfers
; looking at the logic of it...
;
; PRT_PRIO is initialized to 0.
; OK. by default CRT has priority.
;       here, if (uart tx buff is empty) & (prt buff ready to send) and [PRT_PRIO] is set,
;       the printer is given priority, but [PRT_PRIO] gets cleared for the next time through.
; hmm
;       if, during the middle of a partial buffer transfer, a CR is found, it gets set.
; unexpected; this says to give the printer priority the next time we evaluate this again
; I would have thought it would clear it; sending a <CR> is a high latency operation on a printer,
; so the printer should have lower priority for awhile.
;       when we drain the end of a crt/prt buffer, it gets set if
;           (prt buff ready to send) && (we haven't sent a CR since the last time we emptied the crt/prt buffer)
; unexpected; give the printer priority next time through this loop. I guess it prevents printer starvation.
	lxi	h,PRT_PRIO	; 1=give prt traffic priority over crt traffic
	ana	m
	mov	d,a		; d = (uart tx buff empty) & (prt buff ready to send) & [PRT_PRIO]
				; calling this value "old_d" later
	xra	m		; if (uart tx buff empty) & (prt buff ready to send), ...
	mov	m,a		; ... clear the chosen bits in [PRT_PRIO]
	mov	a,d
	cma			; 1=prt buff is empty, or is flow controlled, or uart tx buff isn't empty, or ~[PRT_PRIO]

	lxi	h,CRT_CAN_TX
	ana	m
	mov	e,a		; e = [CRT_CAN_TX] & ~old_d = [CRT_CAN_TX] & (don't use prt buff)
				; TX_POLL looks at 'e'; if the current term bit is 1,
				; it moves crt data, else prt data

	; the following logic is highly convoluted and I'm not sure if I didn't make a mistake,
	; but I've attempted to reduce, logically, what is going on in the comments
	ora	d
	mov	d,a		; d = e | old_d = ([CRT_CAN_TX] & ~old_d) | old_d
				;   = [CRT_CAN_TX] | old_d
	cma			; a = ~([CRT_CAN_TX] | old_d) = ~[CRT_CAN_TX] & ~old_d
	lxi	h,PRT_CAN_TX
	ana	m		; a = [PRT_CAN_TX] & ~[CRT_CAN_TX] & ~old_d
	ora	d		; a = ([PRT_CAN_TX] & ~[CRT_CAN_TX] & ~old_d) | [CRT_CAN_TX] | old_d
				;   = ([PRT_CAN_TX] & ~[CRT_CAN_TX]) | [CRT_CAN_TX] | old_d
				;   = [PRT_CAN_TX] | [CRT_CAN_TX] | old_d
				;   = [PRT_CAN_TX] | [CRT_CAN_TX] | ((uart tx buffer empty) & [PRT_CAN_TX] & [PRT_PRIO])
				;   = [PRT_CAN_TX] | [CRT_CAN_TX] | ((uart tx buffer empty) & [PRT_PRIO])
				; that third term seems to be extraneous
	ana	b		; if tx buffer isn't empty, don't even think about adding more (trumps everything else)

	; the overall effect, convoluted as it is, comes down to:
	; A is the set of terminals to consider in the TX_POLL loop
	;    reg A = ([CRT_CAN_TX] | [PRT_CAN_TX]) & (tx uart buffer is empty)
	;    reg E = map of buffers to read from; 1=crt, 0=prt
	cnz	TX_POLL		; move data from crt or prt buffer to tx buffer for all terms set in A reg

	call	CHECK_HOST	; check if the host cpu needs attention

	in	IN_UART_TXRDY	; only place this is read
	cma
	ani	00FH		; '1' bits indicate which terms are not ready to tx
	mov	d,a
	lda	HAS_UART_TXD	; bitmap of which ports have data in uart tx buffer
	ana	d		; 1=term has buffer data, not flow controlled, and uart is ready to accept a byte
	jz	main5		; jump if no data to send or UART tx buff is full
	mov	d,a
	; fall through ...

	; reg d has a bitmap of which UARTs are ready to accept tx data
	; pick a uart, and transfer one byte each to the corresponding uart.
txd_lookup:
	; this code is similar to TERM_LOOKUP
	rlc			; double the index
	mvi	h,high(TERM_PRIORITY)	; page aligned table
	mov	l,a
	mov	a,m		; grab 1st byte of entry
	mov	c,a
	sta	CUR_UART	; remember which TX port we chose to work on
	out	OUT_UART_SEL
	xra	d		; clear the bit of the chosen terminal
	mov	d,a		; d now contains remaining ports needing attention
	inx	h
	mov	e,m		; grab 2nd byte of entry
	mov	h,e		; term memory page pointer, 20/24/28/2C

	mvi	l,003H		; 2n03 = tx byte count
	dcr	m		; number of bytes left to send
	inx	h		; 2n04
	jnz	main3		; jump if there are more bytes after this one

	; indicate chosen term's uart tx buffer is empty
	lda	HAS_UART_TXD	; bitmap of which ports have data in uart tx buffer
	cma
	ora	c		; clear the bit for this term
	cma
	sta	HAS_UART_TXD

	mov	a,m		; read the tx buffer get pointer
	inr	a		; preincrement
	mov	m,l		; [2n04]=0x04; reset the uart get pointer
	mov	l,a		; hl is the pointer just before it wrapped
	jmp	main4

main3:
	inr	m		; bump pointer [2n04]
	mov	l,m		; hl points to next char
main4:
	mov	a,m		; get next byte to xmit
	out	OUT_UART_DATA	; this is the only spot were data is xmit'd (except cksum_hang)
	xra	a
	ora	d		; d contains map of tx ports needing service
	jnz	txd_lookup	; jump if any other tx ports need servicing
	; fall through ...

main5:
	call	CHECK_HOST

	; keep looping while no active keys ready (KB_BUF_READY),
	; or LINE_REQ is not active,
	; or LINE_REQ is active but LREQ_ACCEPT is false (because 0D, s.f. key, or recall was pressed)
	lxi	h,KB_BUF_READY
	mov	a,m
	inx	h		; LREQ_ACCEPT
	ana	m
	ani	00FH
	jz	MAIN_EVENT	; keep looping while no active keys ready (KB_BUF_READY), or no LINE_REQ is outstanding, or line_req is but L0D43 clared LREQ_ACCEPT
	jmp	RX_POLL		; act on available keystrokes

;------------------------------------------------------------------------------
; When a CBS(0B) command is received, the controller responds with one of the
; following IBS values:
;          00h -- No line request in progress
;          01h -- line request still in progress
;          0Dh -- line request terminated by CR
;          FFh -- Recall key pressed (see note)
;     ENDI(XX) -- S.F. key pressed

; Note on recall:
;     After the FF, the controller may send one or more bytes to the MVP.
;     Each time the MVP sets CPB ready, the controller will send one more data
;     byte with IBS. These are the characters from the entered text, read
;     from right to left, beginning with the cursor position. The beginning of
;     the buffer is indicated by ENDI. This sequence ends whenever the MVP
;     stops setting CPB ready, sends OBS or CBS, or switches address. The
;     controller should not clear the buffer when the 2200 has read all the
;     bytes contained therein. Unfortunately, the 2200 takes some shortcuts
;     for expediency and may reread the buffer later.
;
; Following the query, the MVP may do one of the following:
;    1. Nothing (another query later)
;    2. Delete line request (usually for HALT, and SF keys without parameters)
;    3. Refill -- this is more data to be merged to the present line request,
;       as though the operator typed it (used for recall and DEFFN' quotes).
;       Then End Line Request.
;    4. Terminate Line Request -- used to implement DEFFN' HEX(0D)
;    5. Error Line Request -- this beeps an error and continues the line request
;    6. Ask for data
QUERY_LINE_REQ:
	lda	CUR_TERM
	mov	c,a
	; LREQ_ACTIVE  LREQ_ACCEPT     return     meaning
	; -----------  -----------     ---------  ----------------------------
	;      0            x      ->  00         no line request in progress
	;      1            0      ->  01         line request still in progress
	;      1            1      && last key was 0xFE/0xFF ->  FE/FF  recall key pressed
	;      1            1      && last key was 0x0D      ->  0D     carriage return pressed
	lxi	h,LREQ_ACTIVE	; map of line request in progress
	ana	m
	jz	SEND_A_IBS	; return 00 if no line request in progress
	dcx	h		; LREQ_ACCEPT
	xra	m
	ana	c
	mvi	b,001H
	jz	SEND_B_IBS	; return 01 to indicate line request still in progress
	lhld	CUR_BLOCK	; 2000/2400/2800/2C00
	shld	LINE_REQ_BLOCK
	mov	b,h
	mov	c,m		; 2n00: one-hot encoding of current term

	mvi	l,062H
	mov	e,m		; e = [2n62] (keypress value)
	mvi	a,0FDH
	cmp	e		; 0xFD - e
	dcx	h
	mov	a,m		; a = [2n61] (key prefix byte)
	jnc	RKBE2		; jump if E is <= 0xFD into RELAY_KB_ENTRY code
	; note we are sending either 0xFE or 0xFF, and the docs say 0xFF means recall key pressed
	; FE=dead key, FF=special handling?
	call	SEND_E_IBS

	; see "note on recall" above. send contents of line_req from current cursor position
	; backwards to start of buffer (with that last byte sent with ENDI).
	; that means this code thinks [2n5F/60] is the current position, not [2n69/6A].
	; TBD to reconcile these ideas.
	; de = [2n5F/60] (pointer to end of prefill data)
	mov	b,h
	dcx	h
	mov	e,m
	dcx	h
	mov	d,m

	mvi	l,06BH		; 2n6B: start of line request buffer
qlr1:
	in	IN_2200_STATUS
	xri	0D8H		; 1101 1000
	ani	0FBH		; 1111 1011  (ignore PRIME)
	jnz	qlr3		; jump if not addr 6, or not selected, or CPU is not waiting for input, or CBS, or OBS
	mov	a,e
	xra	l
	jnz	qlr2		; jump if de != hl (lsb check)
	ora	d
	xra	h
	jnz	qlr2		; jump if de != hl (msb check)
	out	OUT_IB9_N	; sending first char, so mark it with ENDI
	jmp	qlr1
qlr2:
	dcx	d		; predecrement pointer (we are reading out in reverse order)
	ldax	d		; get next byte from buffer
	cma
	out	OUT_IB_N	; send it w/o ENDI
	jmp	qlr1
qlr3:
	cpi	008H		; 0000 1000
	jz	qlr1		; jump if port=6, board selected, CPU is not waiting for input, not CBS, and not OBS

	; after recall press, we've sent all the input buffer data to the host
	; (in reverse order)
	; [2n5D/5E] = de = 2n6B (start of buffer) I believe
	mvi	l,05DH
	mov	m,d
	inx	h
	mov	m,e

	; [2n5F/60] = de = 2n6B (start of buffer)
	inx	h
	mov	m,d
	inx	h
	mov	m,e

	; [2n67/68] (last byte used) = de = 2n6B (start of buffer)
	mvi	l,067H
	mov	m,d
	inx	h
	mov	m,e
	jmp	CHECK_HOST	; end of QUERY_LINE_REQUEST command

; port 6 ACCEPT LINE REQUEST DATA, sequence CBS(0C)
; When a CBS(0C) is received after a line request has been completed, the
; controller will send the data. It should only be issued after a query has
; shown that the line is complete.
;
; The controller sends the data if any, then an ENDI as terminator. If the
; ENDI is zero, the line request is complete; if 01h, the controller needs
; more time to finish updating the screen.
ACCEPT_LINE_REQ_DATA:
	lhld	CUR_BLOCK	; 2000/2400/2800/2C00
	shld	LINE_REQ_BLOCK
	mvi	l,067H
	mov	d,m
	inx	h
	mov	e,m		; de = [2n67/68] (last byte used)
	mvi	l,06BH		; 2n6B = start of line_req buffer
	call	DE_SUB_HL	; hl = number of bytes used, de=2n6B
	xchg			; de = number of bytes used, hl=2n6B

	; transfer the contents of the line_req buffer
	mov	b,h
	inr	d		; pre-increment page counter because loop uses pre-decrement to count
	inr	e		; pre-increment byte counter because loop uses pre-decrement to count
	jmp	alrd2
alrd1:
	call	CPU_POLL_WAIT	; wait for host CPU to ask for input
	mov	a,m		; read next line_req buffer byte
	cma
	out	OUT_IB_N	; send it
	inx	h
alrd2:
	dcr	e
	jnz	alrd1
	dcr	d
	jnz	alrd1

	; after transferring contents of buffer, send flag byte with IB9
	call	CPU_POLL_WAIT
	mov	h,b
	mvi	l,063H
	mov	a,m		; [2n63] = line_req flag byte
	mov	c,a
	ani	020H		; test keyin_line_req flag
	mvi	l,067H		; 2n67 (line_req last byte used)
	jz	alrd3
	mvi	l,065H		; 2n65 = (line_req end of buffer)
alrd3:
	mov	d,m
	inx	h
	mov	e,m		; de = [2n65/66] or [2n67/68]

	lda	WANGTERM
	mov	h,b
	mvi	l,000H		; 2n00: this term mask
	ana	m
	jz	alrd4		; jump if not a wang-compatible terminal

	; send <FB><F8> = stop cursor blinking (blinking means input mode)
	mvi	a,0FBH
	rst	5
	mvi	a,0F8H
	rst	5

	mov	a,c		; restore line_req flag byte
	rlc
	jc	alrd5		; jump if underline bit set
alrd4:
	mvi	a,0FFH
	out	OUT_IB9_N	; 00 w/ENDI means the request is complete
	call	CURS_TO_DE
	; fall through ...

SEND_CR_LF:
	mvi	a,00DH		; carriage return
	rst	5		; save it in the terminal output buffer
	mvi	a,00AH		; linefeed
	rst	5		; save it in the terminal output buffer
	mov	h,b
	mvi	l,000H		; 2n00: this term mask
	mov	c,m
	call	TEST_CRT_BUF_ACTIVE
	ei
	; fall through ...

; port 6 DELETE CURRENT LINE REQUEST CBS(03)
; This command causes a pending line request and input buffers of the current
; terminal to be cleared (used at HALT and with special function keys).
DELETE_LINE_REQ:
	mov	a,c		; C=one-hot map of current terminal
	cma
	mov	d,a		; D=A=one cold map of current terminal
	lxi	h,LREQ_ACCEPT
	ana	m		; clear bit corresponding to current term
	mov	m,a
	inx	h		; LREQ_ACTIVE
	mov	a,m
	ana	d		; clear LREQ_ACTIVE bit
	mov	m,a
	; fall through ...

; clear the line_req flag byte
CLEAR_LR_FLAGS:
	mov	h,b
	mvi	l,063H		; 2n63 = line_req flag byte
	mvi	m,000H		; clear flags
	ret

; more of ACCEPT_LINE_REQ_DATA command.
; the command is complete and the data returned to the host, but the request
; used underlines. redraw the buffer without the underscores, and clear any
; underscores which came after the end of the input.
alrd5:
	mvi	l,063H		; 2n63: line_req flag byte
	cmc			; we jumped to alrd5 because carry was set, so this clears it
	rar			; effectely clears the msb (preceding code did rlc)
	mov	m,a		; [2n63] = clear underline bit

	; [2n5F/60](redraw end) = de = [2n65/66](end of buffer) or [2n67/68](last byte used)
	; depending on keyin_line_req flag (see alrd2)
	mvi	l,05FH
	mov	m,d
	inx	h
	mov	m,e

	; hl = [2n67/68](last byte used)
	mvi	l,067H
	mov	a,m
	inx	h
	mov	l,m
	mov	h,a

	; loop filling (hl++) with spaces until hl=de
	; if  keyin_line_req, fill with spaces from (last byte used) to (end of buffer)
	; if !keyin_line_req, fill with spaces from (last byte used) to (last byte used)
	mvi	c,' '		; fill with spaces
	dcx	h		; predecrement because the loop preincrements
alrd6:
	inx	h
	mov	m,c
	mov	a,l
	xra	e
	jnz	alrd6
	ora	h
	cmp	d
	jnz	alrd6

	; rewrite the contents of the buffer, without underline
	call	CURS_TO_SOL	; cursor to start of line
	xra	a
	sta	ULINE_FLAG	; no underline
	call	ALRD_REDRAW	; copy chars from line_req buffer to crt buffer, from sol to (redraw end)
				; ALRD_REDRAW bumps the [2n5C] continuation byte if it completed all work

	; clear any continuations, but if one was pending,
	; signal not everything has been redrawn yet
	mov	h,b
	mvi	l,05CH
	xra	a
	cmp	m
	mov	m,a		; [2n5C] = 00 ; no continuation
	jnz	alrd7		; jump if ALRD_REDRAW completed its work

	; schedule ALRD_REDRAW to be called later to complete the remaining work
	mvi	a,0FEH
	out	OUT_IB9_N	; 01 w/ENDI means the request needs more time to redraw
	mvi	m,06BH		; [2n5C](continuation byte) = 0x6B -> INDTAB 00D6 entry DO_REDRAW

	mvi	l,000H
	mov	c,m		; c=[2n00] (this term mask)
	call	UPDATE_KB_BUF_READY

	; set LREQ_ACCEPT to ensure the ALRD_REDRAW gets called
	inx	h		; LREQ_ACCEPT
	di
	mov	a,m
	ora	c
	mov	m,a		; set "this term" bit in [LREQ_ACCEPT]

	call	TEST_CRT_BUF_ACTIVE	; conditionally set CRT_BUF_ACTIVE bit for the current xterm
	ei
	ret

alrd7:
	mvi	a,0FFH
	out	OUT_IB9_N	; 00 w/ENDI means the request is complete
	jmp	SEND_CR_LF

;------------------------------------------------------------------------------
; EDIT subroutine: move cursor to start of line (non-destructive)
CURS_TO_SOL:		; depth=2
	; hl = 2n6B = start of line_req buffer
	mov	h,b
	mvi	l,06BH
	; fall through ...

; EDIT subroutine: move cursor to the buffer location in HL
CURS_TO_HL:		; depth=2
	xchg
	; fall through ...

; EDIT subroutine: move cursor to the buffer location in DE
CURS_TO_DE:		; depth=2 (call here, rst 5)
	mov	h,b
	mvi	l,06AH		; 2n6A
	mov	a,e		; note: pointers are big endian order
	sub	m
	mov	m,e
	mov	e,a
	dcx	h		; 2n69
	mov	a,d
	sbb	m
	mov	m,d
	mov	d,a
	; here: [2n69/6A]=updated cursor; de = new_location - prev_location
CTDE1:
	mov	h,b
	mvi	l,064H		; 2n64: line_req cursor column
	mov	l,m
	xchg			; hl=cursor delta; e=starting cursor column
	xra	a
	ora	h
	jm	ctde6		; negative delta: cursor moves (logically) left
	mov	a,l
	jz	ctde2		; jump if delta<256
	; h was 1, meaning delta of 256 plus whatever was in L
	; 256 = 3*80+16, which means go down three rows and right 16
	adi	010H		; add 16
	mvi	h,3		; counts required linefeeds
ctde2:
	inr	h		; one more line feed needed
	sui	80		; reduce delta by 80
	jnc	ctde2
	adi	80		; restore quotient/remainder after going one too far
	dcr	h
	mov	l,a		; remainder
	add	e		; add original cursor column
	mov	e,a		; e = new cursor column
	sui	80
	jc	ctde3		; didn't wrap past end of line
	inr	h		; need one more linefeed
	mov	e,a		; e=e-80 (or e mod 80, i.e. wrap it)
ctde3:
	inr	h		; bump by one to compensate for initial dcr at ctde5
	jmp	ctde5

ctde4:
	mvi	a,00AH		; linefeed
	rst	5		; send to terminal output buffer
ctde5:
	dcr	h
	jnz	ctde4
	jmp	ctde10		; horiz cursor motion to target column

; negative cursor delta; logically moving the cursor left.
; this closely follows the logic of the move right code above.
ctde6:
	inr	h		; delta += 256?
	mov	a,l
	jz	ctde7		; jump if H was 0xFF
	; moving left 256 plus whatever is in L
	mvi	h,3		; three lines up...
	adi	0F0H		; and 16 chars left (-256 = -3*80 - 16)
ctde7:
	inr	h
	adi	80
	jnc	ctde7
	dcr	h
	mov	l,a
	add	e
	mov	e,a
	sui	80
	jnc	ctde8
	mov	a,e
	inr	h
ctde8:
	mov	e,a
	mov	a,h
	ana	a
	jz	ctde10		; horiz cursor motion to target column
ctde9:
	mvi	a,00CH		; cursor up
	rst	5		; save in the terminal output buffer
	dcr	h		; prev page
	jnz	ctde9
	; fall through ...

	; adjust the horizontal cursor position right to the target position
ctde10:
	mov	a,l
	cpi	75
	jnc	ctde11		; large cursor motion
	ana	a
	jz	ctde13		; cursor already in the correct column
	mvi	a,0FBH		; send a char compression token
	rst	5		; push compression sequence token
	mov	a,l
	rst	5		; push repetition count
	mvi	a,009H		; nondestructive space
	rst	5		; save cursor motion in the terminal output buffer
	jmp	ctde13

	; moving >75 chars; instead move cursor left and wrap to the right side
ctde11:
	adi	0B0H		; -80
	mov	l,a
ctde12:
	mvi	a,008H		; move to the right by moving left and wrapping
	rst	5		; save in the terminal output buffer
	inr	l
	jnz	ctde12
ctde13:
	mov	h,b
	mvi	l,064H		; 2n64: line_req cursor column
	mov	m,e
	ret

;------------------------------------------------------------------------------
; This code block copies data from the print or crt buffer of the terminal to
; the uart tx buffer for each terminal. The entry point is TX_POLL.

txp1:
	call	CR_LIMIT_XFER	; clamp xfer count if a CR is in the xfer window
				; set the cur term bits in [PRT_PRIO] and [SENT_CR] if CR is found
txp2:
	call	MOV2TXBUF	; move data from crt/print buffer to tx buffer
txp3:
	mvi	l,0FFH		; hl = (22FF|23FF)+n*0x400
	mov	m,e		; update get pointer
	lda	PENDING_TERMS	; retrieve map of which terms still need servicing
	ana	a
	rz			; return when no terms left to service
	lhld	TX_SERVICE	; we don't care about h; l=xfer crt/printer data map
	mov	e,l		; move it to e, where TX_POLL requires it
	; fall through ...

; This is the entry point to this section of code, from MAIN_EVENT.
; entry:
;    a = a bitmap of which terminals we should attempt to service
;    e = term bitmap; 0=send crt data, 1=send printer data
; for each terminal in the A map, up to 16 bytes will be transferred from the
; chosen crt or prt buffer to the corresonding uart tx buffer.
TX_POLL:			; depth=2 (call here, eg: call to TERM_LOOKUP)
	call	TERM_LOOKUP	; c=chosen port bitmap, b=h=page0 ptr, a=remaining ports
	sta	PENDING_TERMS	; which terms are waiting to xfer data to their tx buffer
	mov	d,h		; 20/24/28/2C
	mov	l,e		; bitmap indicating crt/prt data to xfer
	shld	TX_SERVICE	; only place this is written! lsb=crt/prt map, msb=page0 of chosen term

	mvi	l,003H		; 2n03
	shld	CUR_TX_BUF	; points to term's ram block

	; will transfer data to the tx buffer shortly; mark it non-empty
	lda	HAS_UART_TXD	; bitmap of which ports have data in uart tx buffer
	ora	c
	sta	HAS_UART_TXD	; this uart tx buffer has data to send

	; if the target (1=crt/0=printer) has changed, inject routing bytes.
	; note: in e,                   target is 1=crt/0=printer
	;       in TX_RECEIVER, current target is 0=crt/1=printer
	;       that's why (xra e/jz txp8) uses jz. a bit confusing, but it works.
	lda	TX_RECEIVER
	xra	e
	ana	c		; isolate this terminal
	jz	txp8

	; reset the uart tx buffer
	mvi	m,000H		; [2n03] = 0x00
	inx	h
	mov	m,l		; [2n04] = 0x04 (get pointer runs one behind next byte to fetch)

	mvi	b,010H		; limit of the number of bytes to tranfer to uart tx buffer
	mov	a,e
	ana	c
	jz	txp10		; jump to printer path

	inr	h		; 0x21xx+n*0x400 to indicate crt path (LIMIT_TX_XFER does h+=2)
	call	LIMIT_TX_XFER	; de=get ptr, b=byte count to transfer
	jc	txp1		; jump if this doesn't empty the source buffer
	; fall through ...

	; this transfer will empty the remaining bytes of the crt or ptr
	; buffer, so change state related to that event
txp4:
	lda	PRT_CAN_TX
	ana	c
	mov	l,a		; l = nonzero if prt buff has data and isn't flow-controlled

	; see the description of PRT_PRIO where it is declared.
	lda	SENT_CR		; 1=a <CR> has been sent recently
	cma
	ana	l		; a = ~(CR has been sent recently) & (this_prt_can_tx)
	; if A is non-zero, set [PRT_PRIO], else leave [PRT_PRIO] alone. always set [SENT_CR]
	call	S0510		; returns with hl=SENT_CR, a=[SENT_CR]
	xra	c
	mov	m,a		; [SENT_CR] clear bit corrsponding to chosen term

	lxi	h,CRT_BUF_ACTIVE
	call	MARK_BUF_EMPTY
	jz	txp5		; jump if CRT_STAT_PEND bit for this term is false

	; clear the CRT_STAT_PEND bit corresponding to this terminal
	xra	m
	mov	m,a

	lhld	TX_SERVICE	; fetch page0 of term we are servicing for tx
	mvi	a,002H		; crt buffer became empty (see STATUS_RESPONSE)
	call	SET_STATUS_RESP
txp5:
	lhld	TX_SERVICE	; fetch page0 of term we are servicing for tx
	mvi	l,001H		; 2n01: per-term rbi status
	mov	a,m
	ani	0AFH		; 10101111: addr 5&7 are ready
	mov	m,a
	ei
	lda	CUR_TERM
	cmp	c
	jnz	txp6		; jump if this isn't the currently selected term
	di
	lda	SHADOW_RBI
	ani	0AFH		; addr 5&7 are ready
	rst	2		; save A to SHADOW_RBI, send it to !RBI register, enable interrupts
txp6:
	call	MOV2TXBUF	; move crt data to tx buff
	mvi	e,000H		; lsb of start of crt buffer (0x2300+n*0x400)
	; fall through ...

txp7:
	mvi	l,0FEH		; (22FE/23FE) depending how we arrived here
	mov	m,e		; update put pointer
	jmp	txp3

	; inject the crt/printer routing prefix, as the target has changed
txp8:
	; flip the byte stream target, crt <-> printer
	lda	TX_RECEIVER
	xra	c
	sta	TX_RECEIVER	; only place this is written (except the init to 00 on reset)

	; initialize the tx buffer
	mvi	m,002H		; [2n03] = 2 = bytes in tx buffer (account for <FB><nn> routing sequence)
	mvi	b,00EH		; we used up two bytes, so allow only 14 bytes to prevent overflow of 16b uart tx buffer
	inx	h		; 2n04
	mov	m,l		; [2n04]=04 (start of tx buffer-1)
	inx	h

	; before sending a chunk ('b' bytes) of data to the terminal,
	; prefix it with a <FB><xx> routing control sequence
	mvi	m,0FBH
	inx	h

	; jump if the target is the printer (0); drop through if crt (1)
	mov	a,e
	ana	c
	jz	txp9

	mvi	m,0F0H		; the rest of <FB><F0> "route data to crt" sequence
	inr	h		; bump to 0x21xx+n*0x400 to get to crt zone in LIMIT_TX_XFER
	call	LIMIT_TX_XFER	; de=get ptr, b=byte count to transfer
	jc	txp2		; jump if not all data has been transferred
	jmp	txp4

; //// send printer data to terminal
; 'B' is the number of bytes of data to send to the uart.
txp9:
	mvi	m,0F1H		; the rest of <FB><F1> "route data to printer" sequence
txp10:
	mov	a,b
	sui	008H		; the printer gets to send smaller bursts
	mov	b,a		; (8 bytes normally, 6 bytes if routing bytes preceded)
	call	LIMIT_TX_XFER	; de=get ptr, b=byte count to transfer
	jc	txp2		; jump if not all data has been transferred

	lxi	h,PRT_BUF_ACTIVE
	call	MARK_BUF_EMPTY
	jz	txp11		; jump if PRT_STAT_PEND bit for this term is false

	; clear the PRT_STAT_PEND bit corresponding to this terminal
	xra	m
	mov	m,a

	lhld	TX_SERVICE	; fetch page0 of term we are servicing for tx
	mvi	a,001H		; print buffer became empty (see STATUS_RESPONSE)
	call	SET_STATUS_RESP
txp11:
	lhld	TX_SERVICE	; fetch page0 of term we are servicing for tx
	mvi	l,001H		; 2n01: per-term rbi status
	mov	a,m
	ani	0F7H		; addr 4 ready
	mov	m,a
	ei
	lda	CUR_TERM
	cmp	c
	jnz	txp12		; jump if this term isn't currently addressed
	; we just serviced the printer for this term, so indicate there is room
	; to accept more print data
	di
	lda	SHADOW_RBI
	ani	0F7H		; addr 4 is ready
	rst	2		; save A to SHADOW_RBI, send it to !RBI register, enable interrupts
txp12:
	call	MOV2TXBUF	; move prt data to tx buffer
	mvi	e,066H		; start of print buffer (0x2266+n*0x400)
	jmp	txp7

;------------------------------------------------------------------------------
; this routine fetches the printer or crt get pointer into de.
; it reduces the transfer size limit value 'b' passed in if the remaining
; number of bytes in the crt or print buffer is less than the limit.
; entry:
;    h=20/24/28/2C if printer, 21/25/29/2D if crt
;    b=max number of bytes to transfer
; exit:
;    de = printer or crt character get pointer
;    b  = max(b, put-get)
;    cy=1 indicates limited, meaning data is left, 0=b is all the data that's left
LIMIT_TX_XFER:			; depth=1
	; de = printer/crt get pointer (22|23)FF+n*0x400
	inr	h
	inr	h
	mvi	l,0FFH
	mov	d,h
	mov	e,m

	; return with b = max(b, put-get)
	dcx	h		; 0x22FE+n*0x400 (printer put pointer)
	mov	a,m		; h,a = printer put pointer
	sub	e
	mov	l,a		; l = put - get
	mov	a,b		; 0x10, 0x10-8=0x08, 0x0E, 0x0E-8=0x06 depending on caller
	cmp	l
	rc			; return with b=limit if (diff > limit)
	mov	b,l
	ret			; return with b=(put-get)

;------------------------------------------------------------------------------
; If the printer is flow controlled or empty (which means we don't need to mux
; the tx serial stream between printer and crt traffic), return immediately.
; Otherwise, scan from B-1 bytes after get pointer back to the get pointer,
; looking for a <CR>. If one is found, return with B clamped to that character.
;
; The point of this is to make the flow of characters to the crt a bit smoother
; when muxing printer/crt data. Because a <CR> is a high latency operation for
; mechanical printers, there is no pressing need to send the bytes after a <CR>.
; They can wait a while, allowing crt traffic to resume sooner.
;
; entry:
;     b = number of bytes to check
;     c = 1-hot encoding of terminal under consideration
;    de = get pointer of crt or print buffer
; exit:
;     c, de are unaffected
; exit if no CR found:
;     b is unaffected
; exit if CR found:
;     b = number of bytes clamped to the last <CR> found in the span,
;     and the term bits at PRT_PRIO/SENT_CR are set
CR_LIMIT_XFER:			; depth=1
	lda	PRT_CAN_TX
	ana	c
	rz
	mov	h,d
	mov	a,e
	add	b
	mov	l,a		; hl now points to 'b' bytes after get pointer
	mov	d,b		; save original count
	; scan (hl--) looking for 0x0D character, quitting if 'b' bytes are tested
crlx1:
	dcx	h
	mov	a,m
	cpi	00DH
	jz	crlx2		; jump if CR found
	dcr	b
	jnz	crlx1
	mov	b,d		; restore entry value
	mov	d,h		; restore entry value
	ret
crlx2:
	; we found a CR
	mov	d,h		; restore entry value
	inr	a		; a=0x0E, but probably done to clear z flag
	; set both flags after fall through ...

;------------------------------------------------------------------------------
; if z flag is clear, [PRT_PRIO] |= C  (1-hot term map)
; unconditionally,    [SENT_CR]  |= C  (1-hot term map)
S0510:				; depth=1
	lxi	h,PRT_PRIO
	jz	t0519
	mov	a,m
	ora	c		; set bit for current terminal
	mov	m,a
t0519:
	inx	h		; SENT_CR
	mov	a,m
	ora	c		; set bit for current terminal
	mov	m,a
	ret

;------------------------------------------------------------------------------
; Move 'b' bytes from the get pointer (de) to the uart tx buffer.
; The buffer might already have valid bytes, but it is the caller's
; responsibility to not overflow the buffer.
; entry:
;    b = number of bytes to transfer
;   de = get pointer
; exit:
;   de = updated get pointer
;    h = 0x22/23 if printer/crt
MOV2TXBUF:			; depth=1
	lhld	CUR_TX_BUF	; hl = 2n03 of term being serviced by tx polling
	mov	c,m		; c = [2n03]=number of valid bytes in the tx buffer
	mov	a,c
	add	b
	mov	m,a		; [2n03] += b; the buffer now has (old_count + b) valid bytes
	inx	h
	mov	a,m		; a = [2n04]=tx buffer get pointer
	add	c
	mov	l,a		; hl points to the last valid byte

mtxb2:	; block copy loop from (de) -> (hl+1) for 'b' bytes
	ldax	d
	inx	d		; << note: get ptr is postincremented
	inx	h		; << note: put ptr is preincremented
	mov	m,a
	dcr	b
	jnz	mtxb2

	mov	h,d		; hl = 0x22nn or 0x23nn
	ret

;------------------------------------------------------------------------------
; this routine changes the state of a buffer indicating it is empty.
; entry:
;     hl is either CRT_BUF_ACTIVE or PRT_BUF_ACTIVE
;     c is a 1-hot bitmap represting the current terminal
; exit:
;     z flag is cleared if this term's STAT_PEND bit is true
;     disables interrupts
MARK_BUF_EMPTY:		; depth=1
	di
	; clear bit for current term in (CRT/PRT)_BUF_ACTIVE
	mov	a,c
	cma
	ana	m
	mov	m,a
	; clear bit for current term in (CRT/PRT)_CAN_TX
	dcx	h
	mov	a,c
	cma
	ana	m
	mov	m,a
	; test bit for current term in (CRT/PRT)_STAT_PEND
	dcx	h
	mov	a,m
	ana	c
	ret

; respond to polling at addr 2 (STATUS).
; the response sequence is:
;    1. RESET flags
;    2. HALT flags
;    3. DSR flags (manual says MVP returns clock info here)
;    4. terminal 1 status
;    5. terminal 2 status
;    6. terminal 3 status
;    7. terminal 4 status
; the status bytes are encoded as:
;    01: print buffer became empty (requested)
;    02: CRT buffer became empty (requested)
;    04: keyboard buffer not empty or Line Request complete
;    08: unused (MVP uses it)
;    10: unused
;    20: unused
;    40: unused
;    80: unused (MVP uses it)
STATUS_RESPONSE:
	di
	lda	SHADOW_RBI
	ori	002H		; addr 2 is not ready
	rst	1		; save A to SHADOW_RBI, then send it to !RBI register
	lxi	d,00400H	; step between memory block for each term
	lxi	h,HALT_MAP
	mov	c,m		; C=halt map
	mov	m,e		; zero out HALT_MAP
	dcx	h
	mov	b,m		; B=reset map
	mov	m,e		; zero out RESET_MAP
	ei
	call	SEND_B_IBS	; return reset map
	mov	b,c
	call	SEND_B_IBS	; return halt map
; build next response byte
	mvi	b,000H		; next response byte we are building
	mvi	a,008H		; work from term 4 down to term 1
srs1:
	mov	c,a
	sta	CUR_UART
	out	OUT_UART_SEL
	in	IN_UART_STATUS
	ral			; test data set ready bit
	jnc	srs2
	mov	a,b
	ora	c
	mov	b,a
srs2:
	mov	a,c
	rrc
	jnc	srs1		; loop over remaining terminals
	call	SEND_B_IBS	; next status byte (DSR bitmap)

; build next four response bytes
	lxi	h,02002H	; 2n02 = terminal status response byte (2002 for terminal 1)
	mvi	c,004H		; iterate over four terminals
srs3:
	mov	b,m		; fetch [0x2002+n*0x0400]
	mov	m,e		; zero it out
	call	SEND_B_IBS	; send status byte
	dad	d		; move to next block
	dcr	c
	jnz	srs3		; loop over all terminals
	; fall through ...

; send sequence terminating byte 0xF0 w/IB9 set
SEND_F0_IB9:
	mvi	a,0F0H
	; fall through ...

; send sequence terminating byte (in A) w/IB9 set
SEND_A_IB9:
	push	psw
	call	CPU_POLL_WAIT
	pop	psw
	cma
	out	OUT_IB9_N
	ret

; received CBS or OBS with addr=6 or 7
ADDR6_OR_7:
	jnz	ADDR7		; jump if obs/cbs addr=7

	; CBS or OBS with addr=6
	; double check that it was a CBS, not OBS,
	; and reject any CBS which is not 0xFF, or >0x12
	mvi	a,0FFH
	sta	MVP_MODE	; indicate MVP mode
	in	IN_2200_STATUS
	rrc			; test OBS strobe latch
	in	IN_OBUS_N	; read OB_N, and clear OBS/CBS latches
	rc			; return if OBS was set
	ana	a
	jz	SET_ACT_TERM	; jump if OB_N=0 (OBS=FF)
	cpi	0EDH
	rc			; return if OB_N < xED (OB >= 0x12)
	; fall through for io6 CBS() command dispatch ...

; OB    OB_N    command
; --    ----    ----------------------
; FF -> 00   -> select terminal
; 00 -> FF   -> null
; 01 -> FE   -> power-on sequence
; 02 -> FD   -> init current term
; 03 -> FC   -> delete line request
; 04 -> FB   -> keyboard ready check
; 05 -> FA   -> keyin poll request
; 06 -> F9   -> keyin line request
; 07 -> F8   -> line request
; 08 -> F7   -> prefill line request
; 09 -> F6   -> refill line request
; 0A -> F5   -> end of line request
; 0B -> F4   -> query line request
; 0C -> F3   -> accept line request data
; 0D -> F2   -> request crt buffer
; 0E -> F1   -> request print buffer
; 0F -> F0   -> error line request
; 10 -> EF   -> terminal line request
; 11 -> EE   -> device status ** (triple controller and mxe only)
; 12 -> ED   -> start clock ** (triple controller and mxe only)

; A is doubled, then a 16 bit pointer is fetched into hl
; at that offset on page 0, and an indirect branch is taken.
IDXJMPA:
	ral
	mov	l,a
	mvi	h,000H
	mov	d,m
	inx	h
	mov	h,m
	mov	l,d
	pchl	;INFO: index jump

; Set Active Terminal: expected sequence CBS(FF) OBS(nn)
; nn=00 means terminal 1, 01 means terminal 2, etc.
; this routine sets CUR_TERM, which is 0x01, 0x02, 0x04, or 0x08 (one hot encoding).
; it also sets CUR_BLOCK_PAGE with 0x20, 0x24, 0x28, or 0x2C.
SET_ACT_TERM:
	rst	3		; get an OBS byte
	ani	007H
	rlc
	rlc
	adi	020H		; 0->20, 1->24, 2->28, 3->2C
	sta	CUR_BLOCK_PAGE	; this is one of two places where this is set
	mov	h,a
	mvi	l,000H		; hl = 2n00
	mov	a,m
	sta	CUR_TERM	; this is one of two places where this is set
	; fall through ...

; read the term-specific RBI byte at [2n01]. if either bit for addr 4&7 are ready,
; make the actual port bits ready, otherwise leave the RBI state alone.
S05BF:				; depth=1
	mvi	l,001H		; hl = 2n01
	mov	a,m
	ori	0A7H		; bits for addr 1,2,3,6 should be unaffected by "ana c" below
	mov	c,a
	di
	lda	SHADOW_RBI
	ori	048H		; assume addr 4&7 (prt and crt) are busy
	ana	c		; clear addr 4&7 if [2n01] byte has these not busy, but leave ...
	sta	SHADOW_RBI	; ... them clear if the SHADOW_RBI bits had them not busy already
	out	OUT_RBI
	ei
	ret

; This command causes the controller to check the CRT buffer of the current
; terminal. If it is empty, the appropriate status bit is set (address 02h =
; ready) to signal the fact. If not, then the controller will set the bit
; when the buffer does go empty.
REQ_CRT_BUFFER:
	di
	lda	CUR_TERM
	mov	b,a
	lda	CRT_BUF_ACTIVE	; 1=crt buffer is not empty
	ana	b
	lxi	d,0AD02H	; really, D=0xAD=10101101, E=0x02
	jz	rpb1		; jump if buffer is empty; set status response
	lxi	h,CRT_STAT_PEND	; buffer isn't empty; set the status when it is
rcb1:
	mov	a,m
	ora	b		; set bit corresponding to the current term
	mov	m,a
	ei
	ret

; This is just like REQ_CRT_BUFFER, except it refers to the current terminal's
; PRINT buffer, not the CRT buffer.
REQ_PRINT_BUFFER:
	di
	lda	CUR_TERM
	mov	b,a
	lda	PRT_BUF_ACTIVE	; PRINT buffer status bitmap?
	ana	b
	lxi	h,PRT_STAT_PEND
	jnz	rcb1		; jump if buffer isn't empty; set status when it is
	lxi	d,0F501H	; really, D=0xF5=111110101, E=0x01
rpb1:	; CRT/PRINT buffer was queried and is empty; update corresponding status
	di			; not sure why; intr is already disabled
	lda	SHADOW_RBI
	ana	d		; CRT: addr 2&5&7 are ready; PRINT: addr 2&4 are ready
	rst	1		; save A to SHADOW_RBI, then send it to !RBI register
	lda	CUR_BLOCK_PAGE	; page0 of current term
	mov	h,a
	mvi	l,001H		; 2n01 -- per-term RBI status?
	mov	a,m
	ana	d
	mov	m,a
	inx	h		; 2n02 = terminal status response byte
	mov	b,e		; CRT: B=02; PRINT: B=01
	jmp	rcb1

; expected sequence: CBS(07), OBS(XXXXYYZZ), where
;     XXXX: field size (up to 480=0x1E0 bytes according to the manual, but 0x1E8 below)
;     YY:   0x80=underline flag, 0x04=EDIT mode, 0x01=flush old chars
;     ZZ:   CRT cursor column (it should already be there)
LINE_REQ:
	lhld	CUR_BLOCK	; 2000/2400/2800/2C00
	mov	c,m		; this routine doesn't appear to use 'c'
	mvi	l,05CH		; 2n5C
	mvi	b,00FH		; 15 bytes
	call	ZFILL_HL_B	; clear 2n5C - 2n6A (line_req flags and pointers)
	mov	b,h		; 20/24/28/2C
	mov	d,h
	mov	e,l		; 0x6B  (0x5C+15)

	; [2n69/6A] (cursor location) = 2n6B (start of LINE_REQ buffer)
	dcx	h
	mov	m,e		; 2n6a = 6b
	dcx	h
	mov	m,d		; 2n69 = 2n  -- note: msb is in lower address!

	; [2n67/68] (last byte used) = 2n6B (start of LINE_REQ buffer)
	dcx	h
	mov	m,e
	dcx	h
	mov	m,d

	; [2n5D/5E] (redraw start) = 2n6B (start of LINE_REQ buffer)
	mvi	l,05DH
	mov	m,d
	inx	h
	mov	m,e

	; [2n5F/60] (redraw end)= 2n6B (start of LINE_REQ buffer)
	inx	h
	mov	m,d
	inx	h
	mov	m,e

	; receive two OBS bytes (field size), and compute end of buffer
	rst	3
	mov	d,a
	rst	3
	mov	e,a
	lxi	h,001E8H	; = 488, buffer size limit
	call	MAX_HL_MIN_DE	; de = smaller of 488 and the requested size
	mov	h,b		; 20/24/28/2C
	mvi	l,06BH		; 2n6B is the start of the line request buffer
	dad	d
	xchg			; de = pointer to end of buffer

	; 2n65/66 = pointer to end of LINE_REQ buffer (not sure if inclusive or not)
	mov	h,b
	mvi	l,065H
	mov	m,d
	inx	h
	mov	m,e

	; get one OBS byte for YY, the edit/underline/flush flag byte
	rst	3
	mvi	l,063H		; 2n63 = flag byte; 80=underline, 04=edit, 01=flush previously queued keystrokes
	ani	0ECH		; 11101100b : interestingly clears the 01 bit before testing it and doesn't clear undefined bits
	mov	m,a

	; get one OBS byte for ZZ, the CRT cursor column of the field start
	rst	3		; get an OBS byte (ZZ byte)
	mvi	l,064H		; 2n64 = current column of CRT cursor
	mov	m,a
	ret

lreq1:
	in	IN_OBUS_N
	cma			; right polarity
	mov	m,a		; save it to hl buffer
	mov	a,l
	cmp	e
	jnz	lreq2
	mov	a,h
	cmp	d
	jz	BUFFER_OBS_6_DATA
lreq2:
	inx	h		; bump hl if hl != de
	; fall through ...

; accumulate run of addr 6 OBS data to buffer at hl
; keep accepting data until we get something other than OBS.
; on entry, HL is where we want to write, and DL points to the end of buffer
; if we get more OBSs than buffer expects, last char of the buffer
; gets written over and over without incrementing the put pointer,
; but no error occurs.
; on exit, HL=pointer to one past last byte written (or last byte written if we clamped),
BUFFER_OBS_6_DATA:
	in	IN_2200_STATUS
	rar
	jc	lreq1		; OBS is set
	ani	07DH		; x1111101(x)
	cpi	068H		; 01101000 = no cbs, no prime, not waiting for input, selected, addr=6
	rnz			; return if not true
	jmp	BUFFER_OBS_6_DATA

;------------------------------------------------------------------------------
; The refill command is identical to a prefill except that it does not cause
; repositioning of the cursor to the beginning. Thus the characters are
; treated as keystrokes. It is normally used for RECALL and DEFFN' quotes.
; It is generally followed by an END-OF-LINE-REQUEST CBS(0A) or a
; TERMINATE-LINE-REQUEST CBS(10).
REFILL_LINE_REQ:
	lhld	CUR_BLOCK	; 2000/2400/2800/2C00
	mvi	l,063H		; 2n63: line_req flag byte
	mov	a,m
	ori	010H		; refill_line_req flag
	mov	m,a
	; fall through ...

; expected sequence: CBS(08) OBS(YYYYYY...)
; This optional command can be sent after a line request command, CBS(07),
; to prefill the desired line with the supplied characters YYYY... starting
; with the leftmost position. The characters are treated as keystrokes.
; The cursor is left at the leftmost character. The string of characters is
; terminated by the next CBS, which will normally be an END-OF-LINE-REQUEST.
PREFILL_LINE_REQ:
	lhld	CUR_BLOCK	; 2000/2400/2800/2C00
	mov	b,h
	mov	c,m		; one-hot term encoding

	; de = [2n65/66] = end of line_req buffer
	mvi	l,065H
	mov	d,m
	inx	h
	mov	e,m

	; hl = [2n5F/60] = start of buffer, or where previous prefill ended (probably)
	mvi	l,05FH
	mov	a,m
	inx	h
	mov	l,m
	mov	h,a

	call	BUFFER_OBS_6_DATA	; fill buffer at [hl] up to de with io6 OBS data
	xchg			; de=where prefill ended; hl points to end of buffer
	mov	h,b
	mvi	l,05FH
	mov	m,d
	inx	h
	mov	m,e		; [2n5F/60] <- where prefill ended

	; save last used by pointer: [2n67/68] = larger of [2n67/68] vs [2n5F/60]
	mvi	l,067H
	call	HLPTR_MAXMIN_DE2	; hl <- [2n67/68]; conditionally swap hl/de
	xchg			; de=larger, hl=smaller
	mov	h,b
	mvi	l,067H
	mov	m,d
	inx	h
	mov	m,e
	ret

;------------------------------------------------------------------------------
; This command is used (after optional PREFILL or REFILL) to cause all the same
; actions as the operator pressing EXEC. It is normally used for the BASIC
; statement DEFFN' HEX(0D).
TERMINATE_LINE_REQ:
	mvi	a,065H		; continuation code; INDTAB offset 00CA (CURSOR_ON)
	jmp	END_LINE_REQ

; This command causes the line request to resume, just like END_OF_LINE_REQ,
; except it beeps first. It should not be used in conjunction with PREFILL or
; REFILL. It is normally used for undefined function keys.
ERROR_LINE_REQ:
	mvi	a,064H		; continuation code; INDTAB offset 00C8 (ERROR_LINE_REQ2)
	jmp	END_LINE_REQ

; A special command must be supplied to signal the end of a line request
; sequence which consists of the setup and prefill if desired. The last command
; sent, however, must be a CBS(0A), to signal the microcode to invoke the line
; request. Nothing is sent to the CRT until the CBS(0A) is issued.
;
; This command is also used after successful RECALL or DEFFN' text entry to
; signal the controller to resume processing the line request.
;
; NOTE: The 2200 will sometimes skip this command when it knows it will read in
; the data from the controller on the next command. This is a violation of
; protocol but the controllers live with it. Should the 2200 skip this command
; and go directly to CBS(0C) [ ACCEPT_LINE_REQ_DATA ] the controller should
; update the screen before allowing another Line Request to be started.
END_OF_LINE_REQ:
	mvi	a,05FH		; continuation code; INDTAB offset 00BE (CURSOR_ON)
	; fall through ...

; common to TERMINATE_LINE_REQ, ERROR_LINE_REQ, and END_OF_LINE_REQ
END_LINE_REQ:
	lhld	CUR_BLOCK	; 2000/2400/2800/2C00
	mov	b,h
	mov	c,m		; one-hot term encoding
	call	SAV_CONT	; save continuation (on return hl=KB_BUF_READY)

	; mark [LREQ_ACCEPT] true to cause continuation processing
	inx	h		; LREQ_ACCEPT
	mov	a,m
	ora	c
	mov	m,a

	; mark this term as having an active line req
	inx	h		; LREQ_ACTIVE
	mov	a,m
	ora	c
	mov	m,a
	ret

;------------------------------------------------------------------------------
; On entry A indicates an offset into the INDTAB table. At a later time, this
; offset is recalled and IDXJMPA does a branch through the table. I'm calling
; this is a "continuation" as it continues processing a suspended command.
; The offsets used are shown here. Some continuations increment the [2n5C]
; byte to link to another continuation.
;
; Why interrupt commands like this? Because the supervisor has to multitask
; between four terminals and none can block the others. For instance,
; on END_OF_LINE_REQ, all the stored line buffer contents must be sent to the
; terminal, but we can't afford to wait in one event loop to push all those
; bytes. Instead, we push what fits in the uart tx buffer, then suspend and
; later the routine will continue to send more bytes as room appears in the
; uart tx buffer.
;
;     caller               A       IDXJMPA offset
;     -------------------  ----    --------------
;     END_OF_LINE_REQ      0x5F -> 0xBE
;     EDIT_SF10 (insert)   0x61 -> 0xC2
;     EDIT_SF9  (delete)   0x61 -> 0xC2
;     ERROR_LINE_REQ       0x64 -> 0xC8
;     TERMINATE_LINE_REQ   0x65 -> 0xCA
;
; After saving that index value, UPDATE_KB_BUF_READY is called
SAV_CONT:			; depth=1
	mvi	l,05CH
	mov	m,a		; [2n5C] = continuation byte
	; fall through ...

; update KB_BUF_READY flags by OR'ing in C mask
; C: 01=term 1 has non empty KB buff, .., 08=term 4 has non-empty KB buff
; note: it returns with the updated mask in A and HL=KB_BUF_READY
UPDATE_KB_BUF_READY:		; depth=1
	lxi	h,KB_BUF_READY
	di
	mov	a,m
	ora	c
	mov	m,a
	ei
	ret

;------------------------------------------------------------------------------
; this initializes some data structs for one terminal.
; it is called by both INIT_CUR_TERM and POWER_ON_INIT code paths.
; the CRT buffer is initialized with an init command sequence.
INIT_CRT_BUF:			; depth=2
	call	SET_CRT_BUF_ACTIVE	; we will send an init sequence
	mov	a,c
	sta	CUR_TERM	; this is one of two places where this is set
	call	TERM_LOOKUP
	mov	a,b
	sta	CUR_BLOCK_PAGE	; this is one of two places where this is set
	mvi	l,000H		; 2n00
	shld	CUR_INIT	; save it for use by INIT_PRT_BUF
	mov	m,c		; 2n00: this term 1-hot bitmap

	mvi	l,016H		; 2n16
	mvi	b,005H		; fill five bytes (2n16 through 2n1A)
	call	ZFILL_HL_B	; zero out keyboard rx buffer pointers (2n16-2n1A)
	call	INIT_KB_BUF

	; initialize the crt buffer with this 18 byte sequence:
	;    <FB><F6>   -- immediate command to reset the crt
	;    <01> x 16  -- delay a few character times? <01> means home cursor
	inr	h
	inr	h
	inr	h
	mvi	l,0FFH
	mvi	m,000H		; 23FF+n*0x400=0x00 = crt buffer get pointer
	dcx	h
	mvi	m,012H		; 23FE+n*0x400=0x12 = crt put pointer
	mvi	l,000H
	mvi	m,0FBH		; 2300+n*0x400=0xFB
	inx	h
	mvi	m,0F6H		; 2301+n*0x400=0xF6  << INIT_PRT_BUF overrides this
	inx	h
	mvi	b,010H		; fill 16 bytes (2302 to 2311)+n*0x0400 ...
	mvi	a,001H		; ... with 0x01
	jmp	A_FILL_HL_B	; fill and return

;------------------------------------------------------------------------------
; initialize the keyboard rx buffer.
INIT_KB_BUF:			; depth=1
	mvi	l,01AH		; 2n1A: get pointer of rx buffer
	mov	a,l
	mov	m,a
	dcx	h		; 2n19: put pointer of rx buffer
	mov	m,a
	dcx	h		; 2n18: count of entries in keyboard buffer
	mvi	m,000H
	ret

;------------------------------------------------------------------------------
; INITIALIZE CURRENT TERMINAL CBS(02) OBS(00)
; "This command will cause the CRT screen, pending line request CRT buffer,
;  print buffer, and input buffer of the current terminal to be cleared
;  (used at RESET)."
; Interestingly, if the trailing byte isn't 00, the print tx buffer is not
; reset. That behavior isn't described by 2236MXE_Documentation.8-83.pdf
; Apparently it is sent in response to a RESET keypress in MVP mode, where
; we want to reset the foreground keyboard/terminal without interfering with
; a background print job.
INIT_CUR_TERM:
	lda	CUR_BLOCK_PAGE	; page0 of current term
	mov	h,a
	call	S05BF
	lda	CUR_TERM
	mov	c,a
	cma
	mov	b,a
	mvi	e,10		; there are 10 bitmap flags to clear
	lxi	h,WANGTERM	; starting at this byte
	di
ict1:
	mov	a,m
	ana	b		; clear current term bit
	mov	m,a
	inx	h
	dcr	e
	jnz	ict1		; loop through all 10 bytes

	ei
	call	INIT_CRT_BUF
	ei
	rst	3		; get an OBS byte
	ana	a
	cz	INIT_PRT_BUF	; 00 is normal
	ei
	ret

;------------------------------------------------------------------------------
; initialize the print tx buffer for current terminal
INIT_PRT_BUF:			; depth=2
	lhld	CUR_INIT	; term ram ptr INIT_CRT_BUF saved off
	inr	h
	inr	h
	mvi	l,0FEH		; hl=0x22FE+n*0x0400
	mvi	m,076H		; prt put pointer = 0x76
	inx	h		; hl=0x22FF+n*0x0400
	mvi	m,066H		; prt get pointer = 0x66
	mvi	l,066H		; hl=0x2266+n*0x0400
	mvi	b,010H		; 16 bytes
	call	ZFILL_HL_B	; fill with <00>

	; INIT_CRT_BUF was initialized to send <FD><F6>, which resets the crt.
	; INIT_PRT_BUF is conditionally called after INIT_CRT_BUF.
	; here we replace the <F6> with <F2>, which means "full terminal reset".
	inr	h
	mvi	l,001H		; hl=0x2301+n*0x0400 = 2nd byte of crt tx buff
	mvi	m,0F2H

	; assume flow control is in effect; wait for "PRT ON" flow control
	; before resuming traffic to the printer
	lxi	h,PRT_STOPPED
	mov	a,m
	ora	c
	mov	m,a
	dcx	h		; PRT_BUF_ACTIVE
	jmp	SET_BUF_ACTIVE

;------------------------------------------------------------------------------
; Whenever an io6 CBS(04) is received, the controller checks the keyboard
; buffer of the selected terminal. An IBS(00) is sent if it is empty, and non-
; zero if there is a character (the non-zero byte is of no particular value).
KB_READY_CHECK:			; depth=2
	call	CPU_POLL_WAIT
	lda	CUR_TERM	; one-hot
	mov	c,a
	lxi	h,KB_BUF_READY
	ana	m		; 0=empty, 1=non-empty
	cma
	out	OUT_IB_N
	ret

; Whenever a command code of 06h is received, the control checks the keyboard
; buffer. If there is a byte in it, this command is treated exactly as the
; command 05h (KEYIN_POLL_REQUEST). If not, an IBS(00) is sent, and a special
; Keyin Line Request is set up. An interrupt (completed line request) will be
; generated at address 02h when a character is later received.
KEYIN_LINE_REQ:
	di
	lhld	CUR_BLOCK	; 2000/2400/2800/2C00
	mvi	l,063H		; 2n63: line_req flag byte
	mvi	m,002H		; set flag for KEYIN_LINE_REQ
	ei
	; fall through ...

; Whenever a command code of 05h is received, the controller checks the keyboard
; buffer. If there is no character an IBS(00) is returned. Otherwise, a non-
; zero byte is returned followed by an IBS of the byte (with ENDI if it is a
; Special Function key). It seems that if the user hit one of the atom keys,
; eg PRINT, it returns the atom value and doesn't expand it.
KEYIN_POLL_REQ:
	call	KB_READY_CHECK		; emits IBS(00) or IBS(non-zero)
	rz
	lda	CUR_BLOCK_PAGE		; page0 of current term
	mov	b,a
	call	CPU_POLL_WAIT		; wait for !CPB
	call	RELAY_KB_ENTRY		; this must return the next kb buffer byte via IBS, optionally with IB9 set if it is a special function key
	lda	CUR_BLOCK_PAGE		; page0 of current term
	mov	b,a
	jmp	CLEAR_LR_FLAGS		; clear keyin_line_req bit (02H) of line_req flag byte

;------------------------------------------------------------------------------
; continuation of the CHECK_HOST code
; host is ready for input (!CPB)
CKHOST2:				; depth=4 (call here, jump to CHECK_HOST)
	in	IN_2200_STATUS
	ani	0E0H			; isolate the address
	cpi	040H
	jz	STATUS_RESPONSE		; jump if addr 2
	jnc	CHECK_HOST		; jump if addr 3-7 (ignore them)
	; we know the addr must be 1, VP keyboard, because 0 is never received
	lda	SHADOW_RBI
	rrc
	rc
	lxi	b,02001H		; b=0x20=term 1; c=0x01 one hot encoding
	; fall through ...

;------------------------------------------------------------------------------
; read the next entry of the kb receive buffer, and send it to the host CPU
; possibly with the IB9 bit set (for s.f. keys).
RELAY_KB_ENTRY:			; depth=3
	call	NEXT_KB_PAIR	; on return, A=prefix byte, E=char code
	; fall through ...

; reg A = prefix byte, reg E = key code
RKBE2:
	inr	a
	jz	rkbe4		; jump if a=FF: KB buffer was empty (E=00) or special (eg, backspace, CR)
	inr	a
	jz	rkbe3		; jump if A=0xFE (dead key)
	inr	a		; set z flag if A=0xFD (EndI/Atom)
	mov	a,e
	rlc
	jnz	rkbe6		; jump if A!=0xFD, namely 0xFC
	jc	SEND_E_IBS	; jump if E >= 0x80 (atom)
	; fall through ...

; respond to CPU with byte=E and IB9 set (s.f. key)
SEND_E_IB9:
	mov	a,e
	jmp	SEND_A_IB9	; send terminating byte (in A) w/IB9 bit set

rkbe3:
	mov	h,b
	mvi	l,016H		; 2n16
	mov	m,e		; save key value
	mvi	b,0FFH
	call	UPDATE_KB_BUF_READY
	mov	a,c
	rrc
	jnc	SEND_B_IBS	; jump if not term 1
	lda	SHADOW_RBI
	ani	0FEH		; addr 1 is ready
	rst	1		; save A to SHADOW_RBI, then send it to !RBI register
	jmp	SEND_B_IBS
rkbe4:
	mov	a,e
	cpi	0BDH		; EDIT key
	jz	SEND_F0_IB9	; relay EDIT key event to host
	cpi	0E7H
	jc	SEND_A_IBS	; jump if char < 0xE7
	mov	h,b
	inr	h
	inr	h
	inr	h
	mvi	l,0FEH		; hl=23FE+n*0x0400
	mov	a,m		; put offset
	cmp	l		; don't write past 23FD
	jz	rkbe5		; jump if at end
	inr	m		; bump put pointer
	mov	l,m
	dcr	l		; hl points at one byte before put ptr offset
	mvi	m,007H		; bell character?
	call	SET_CRT_BUF_ACTIVE
rkbe5:
	mvi	b,023H		; '#'
	jmp	SEND_B_IBS
rkbe6:
	jc	SEND_E_IB9	; jump if keycode was >= 0x80
	; fall through ...

;------------------------------------------------------------------------------
; send E to the CPU via an IBS strobe after waiting for the CPU to ask for it
SEND_E_IBS:			; depth=2
	mov	a,e
	; fall through ...

; send A to the CPU via an IBS strobe after waiting for the CPU to ask for it
SEND_A_IBS:			; depth=2
	mov	b,a
	; fall through ...

; send B to the CPU via an IBS strobe after waiting for the CPU to ask for it
SEND_B_IBS:			; depth=2
	call	CPU_POLL_WAIT
	mov	a,b
	cma
	out	OUT_IB_N
	ret

;------------------------------------------------------------------------------
; has the host given us something to do? OBS, CBS, or !CPB (request response)
; this subroutine doesn't return anything (in a reg or in flags).
; it is called simply to poll for something to do.
CHECK_HOST:			; depth=4  (here, call COMPRESS)
	in	IN_2200_STATUS
	ani	00FH		; isolate (SEL & !CPB), PRIME, CBS, OBS status
	rz			; return if the cpu hasn't done anything
	mov	b,a
	ani	003H
	jnz	ckhost3		; jump if CBS or OBS
	ora	b
	ani	004H		; isolate PRIME status
	jz	CKHOST2		; jump if !reset, which means !cpb (host ready for input)
	out	OUT_CLR_PRIME	; clear the reset latch (only places this is done)
	jmp	CHECK_HOST	; poll again

ckhost3:
	; we've received CBS or OBS strobe
	lhld	CUR_BLOCK	; 2000/2400/2800/2C00
	mov	b,h
	mov	c,m		; [2n00] one-hot encoding of this term

	; route to the appropriate addr handler
	in	IN_OBSCBS_ADDR
	ani	0E0H		; isolate the address
	cpi	0C0H
	jnc	ADDR6_OR_7	; jump if addr 6 or 7
	cpi	0A0H
	jc	ADDR1_TO_4	; jump if addr 1-4

	; board is addressed at addr 5 (VP mode CRT)
	call	SET_VP_MODE
	lxi	d,0A050H	; d=0xA0 means addr 5, e=0x50 (addr7|5 rbi bit)
	jmp	ckhost4

; received CBS or OBS with addr=7
ADDR7:
	lxi	d,0E040H	; d=0xE0 means addr 7, e=0x40 (addr7 rbi bit)
	lda	CUR_TERM
	rrc
	jnc	ckhost4		; jump if not term 1
	mvi	e,050H		; e = 0x50 (addr7|5 rbi bit)
	; fall through ...

; received CBS or OBS with addr=5 or 7 (send data to crt)
ckhost4:
	lxi	h,CRT_BUF_ACTIVE
	shld	BUF_ACTIVE_IND
	lxi	h,03FEH		; index to 23FE+n*0x400, CRT buffer get/put pair
	; fall through ...

; board addressed at addr 4 (printer), 5 (vp crt), or 7 (mvp crt)
; and the host CPU has performed either a CBS or OBS.
; hl contains a pointer to the appropriate output buffer (printer or crt)
; d[7:5] contains the address active at the start of the sequence,
;        which is used to check that the same port is still addressed later
; e contains an rbi bitmap of which ports correspond to this address
CKHOST5:
	call	COMPRESS	; keep accumulating bytes, compressing into buffer
	jnz	ckhost7		; jump if ready/busy state hasn't changed

	; update the term-local rbi byte and the active rbi status
	; to indicate the current port(s) are busy
	push	h
	lda	CUR_BLOCK_PAGE	; page0 of current term
	mov	h,a
	mvi	l,001H		; 2n01: per-term rbi status
	di
	mov	a,m
	ora	e		; mark port(s) busy
	mov	m,a
	pop	h
	di			; why di again? di was just 5 instructions ago
	lda	SHADOW_RBI
	ora	e		; mark port(s) busy
	rst	2		; save A to SHADOW_RBI, send it to !RBI register, enable interrupts

	in	IN_2200_STATUS
	rar
	jc	ckhost6		; OBS is set
	rar
	jnc	ckhost7		; CBS is not set

	; CBS to addr 4/5/7
	in	IN_OBSCBS_ADDR	; card address at time of CBS
	ani	0E0H		; isolate address
	cmp	d
	cz	CBS457		; if addr hasn't changed, handle the CBS(nn) data
	jmp	ckhost7

ckhost6:
	; OBS to addr 4/5/7
	in	IN_OBUS_N
	cma
	mov	m,a		; save datum to [hl++]
	inx	h
	cpi	0FBH		; literal FB is escaped as FB D0
	jnz	ckhost7
	mvi	m,0D0H
	inx	h
	; fall through ...

ckhost7:
	mov	e,l
	mvi	l,0FEH		; 22FE/23FE
	mov	a,m		; fetch old put pointer
	mov	m,e		; update put pointer to where hl stopped
	cmp	e
	rz			; return if the put pointer didn't move

	; we saved something, so mark the buffer as non-empty
	lhld	BUF_ACTIVE_IND	; pointer to either (PRT|CRT)_BUF_ACTIVE
	lda	CUR_TERM
	mov	c,a
	call	SET_BUF_ACTIVE
	ei
	ret

;------------------------------------------------------------------------------
; if [23FE+n] != [23FF+n], ...
; set the CRT_BUF_ACTIVE bit (buffer is non-empty) for the current xterm.
; note it doesn't clear the bit if the buffer is empty.
TEST_CRT_BUF_ACTIVE:		; depth=1
	mov	h,b		; 20/24/28/2C
	inr	h
	inr	h
	inr	h
	mvi	l,0FEH		; 23FE+n*0x0400
	mov	a,m
	inx	h		; 23FF+n*0x0400
	cmp	m
	rz			; done if [23FE] = [23FF] (guess: put ptr == get ptr)
	; fall through ...

; set the CRT_BUF_ACTIVE bit for the current xterm
SET_CRT_BUF_ACTIVE:		; depth=1
	lxi	h,CRT_BUF_ACTIVE
	; fall through ...

; set the buffer active bit corresponding to the current terminal.
; the hl pointer should be set either to CRT_BUF_ACTIVE or PRT_BUF_ACTIVE on entry.
SET_BUF_ACTIVE:
	di
	mov	a,m		; map of terms with non-empty crt/prt buffers
	ora	c
	mov	m,a		; mark current term buffer as non-empty
	inx	h		; (CRT/PRT)_STOPPED
	cma			; A: 1=empty
	ora	m		; A: 1=empty or stopped
	cma			; A: 0=empty or stopped; 1=non-empty and not stopped
	dcx	h
	dcx	h
	mov	m,a		; (CRT/PRT)_CAN_TX = ![FOO_STOPPED] && [FOO_BUF_ACTIVE]
	ret

;------------------------------------------------------------------------------
; handle CHECK_HOST CBS or OBS to addr 1, 2, 3, or 4
; for some perverse reason, it isn't inline with the rest of CHECK_HOST code
ADDR1_TO_4:
	cpi	080H
	jc	GET_OBS_DAT	; if addr 1-3, ignore the OBS/CBS

	; OBS to addr 4 sends bytes to the printer
	lxi	h,PRT_BUF_ACTIVE
	shld	BUF_ACTIVE_IND
	lxi	h,002FEH		; index into 0x22FE+n*0x400: PRT buffer get/put pointer pair
	lxi	d,08008H		; d=0x80 means addr 4, e=0x08 (addr4 rbi bit)
	jmp	CKHOST5

;------------------------------------------------------------------------------
; called from CKHOST5 when a the first byte has been sent to addr 4/5/7.
; we save the byte in the output buffer and then stay here waiting for
; more CBS/OBS to the same address. If repeated characters occur, this
; routine converts them to the terminal compression encoding. We fall
; out of this routine if the address changes, or we don't see more data
; quickly enough, or if the buffer wraps.
;
; on entry:
;    hl contains a pointer to the (printer/crt) output buffer offset
;    d[7:5] contains the address active at the start of the sequence, which
;           is used to check that the same port is still addressed later
;    e contains an rbi bitmap of which ports correspond to this address

; on return:
;    the z flag set if the interface must be set to !ready
COMPRESS:			; depth=3 (call here, call TIMED_POLL)
	lda	CUR_BLOCK_PAGE	; page0 of current term
	add	h		; hl is invariably either 0x02FE or 0x03FE
	mov	h,a		; now hl points to 0x22FE or 0x23FE
	mov	l,m		; hl is now the current prt/crt buffer put pointer

	; check if the current terminal supports Wang compression
	lda	WANGTERM
	mov	b,a
	lda	CUR_TERM
	ana	b
	sta	CUR_CAN_CMPRS	; the only place this location is set

	; prepare to look for runs of repeated characters
	mvi	c,0FBH		; escape code
	mov	b,c		; b=previous char (esc care ensures on false positive)
	jmp	cmprs4

	; poll for more OBS data (from the same address in d: 4/5/7)
	; and write it into consecutive addresses at [hl].
	; at this point there is no established repeat sequence.
	; if we detect a repeat, we will back up and overwrite the initial
	; literal characters with a compression sequence.
cmprs1:
	in	IN_2200_STATUS
	rrc
	jnc	cmprs5		; jump if not OBS
	in	IN_OBSCBS_ADDR
	ani	0E0H		; isolate addr
	cmp	d
	rnz			; the address changed, so we are done here
	in	IN_OBUS_N
	cma			; normal polarity
cmprs2:
	mov	m,a		; save next byte
	inx	h
	cmp	c		; c is always 0xFD here
	jz	cmprs8		; branch if literal 0xFD was received
	cmp	b
	jz	cmprs7		; branch if this char is same as last
cmprs3:
	mov	b,a		; new previous char

	; return if the hl buffer pointer needs to wrap.
	; otherwise keep accumulating OBS data
cmprs4:
	mvi	a,0F9H
	cmp	l
	jnc	cmprs1		; keep accumulating if L <= 0xF9
	xra	a		; set z flag: update ready/busy status
	ret			; return to CKHOST5

	; we've been accumulating OBS data in the hl buffer,
	; but we've gotten something other than OBS
cmprs5:
	rrc
	jnc	cmprs6		; jump if not CBS
	in	IN_OBSCBS_ADDR
	ani	0E0H		; isolate address
	cmp	d
	rnz			; the address changed, so we are done here
	call	CBS457		; CBS(nn) requires special handling
	jmp	cmprs13

	; no OBS, no CBS. is current port still addressed?
cmprs6:
	rlc			; \_ restore original IN_2200_STATUS
	rlc			; /  value read
	xri	010H		; flip board select bit polarity
	ani	0F8H		; ignore prime, cbs, obs bits
	cmp	d
	rnz			; return if addr mismatch, or !board_sel, or !CPB
	call	POLL_150	; see if we get a quick CBS or OBS
	rnz			; don't wait any longer if none
	jmp	cmprs1		; process the new byte

	; the start of a repeated sequence has been seen
cmprs7:
	lda	CUR_CAN_CMPRS
	ana	a
	jz	cmprs4		; jump if terminal doesn't support compression
	mvi	c,2		; we've seen the start of a run
	jmp	cmprs11

	; we got a literal <FD> byte. emit <FD><D0> escape sequence.
cmprs8:
	mvi	m,0D0H
	inx	h
	jmp	cmprs3

cmprs9:
	rrc
	jc	cmprs12		; jump if CBS
	call	TIMED_POLL	; time-limited poll for OBS or CBS, 'L' attempts made
	; Bizarrely, the caller doesn't set L; instead it is just whatever the
	; output buffer put pointer happens to be. I guess the idea is that is
	; just isn't unbounded, but the window varies from 50us to 12ms!
	jnz	cmprs12		; didn't get a CBS or OBS in a timely fashion
cmprs10:
	in	IN_2200_STATUS
	rrc
	jnc	cmprs9		; jump if not OBS
	in	IN_OBSCBS_ADDR
	ani	0E0H		; isolate addr
	cmp	d
	rnz			; return if addr has changed
	in	IN_OBUS_N
	cma			; normal polarity
	cmp	b
	jnz	cmprs14		; jump if it ends the repeat
	inr	c		; repeat_count++
	mvi	a,79
	cmp	c
	jc	cmprs12		; jump if count >= 80
	; fall through ...

	; we have started a run of identical characters (two or more)
	; the repetition count is in c, and the prev char is in b.
cmprs11:
	mvi	a,0F9H
	cmp	l
	jnc	cmprs10		; keep accumulating if pointer <= 0x23F9+n*0x400
	; fall through ...

	; end the repeated char sequence at this point for one of these reasons:
	; (1) we got a CBS (ending the string of OBS data)
	; (2) we didn't see a CBS or OBS in a timely manner
	; (3) the repeat count is 80 characters
cmprs12:
	mvi	a,2
	cmp	c
	cc	CMPRS15		; rewrite with compression if count > 2
cmprs13:
	mvi	c,0FBH		; anticipate looking for literal <FB> later
	mov	b,c		; prev char=<FB> prevents matching
	jmp	cmprs4

	; received a character that doesn't continue a repeat string.
	; a=new char, b=prev char, c=how many repeats
cmprs14:
	push	psw		; depth
	mvi	a,2
	cmp	c
	cc	CMPRS15		; if count > 2, rewrite first two inline repeated characters with compression
				; otherwise let literal inline bytes stay where they are
	mvi	c,0FBH		; anticipate looking for literal <FB> later
	mov	b,c		; prev char=<FB> prevents matching
	pop	psw
	jmp	cmprs2		; keep receiving OBS bytes, and start a new sequence perhaps

; two bytes literal bytes were written into the buffer before detecting the sequence.
; back up two bytes and overwrite with the compression sequence (either space or char)
; b=character being sent, c=number of times
;
; as literal bytes are saved the buffer, hl is simply incremented up to, and
; including, 0x23F9. But the three byte sequence can write 0x23F8/F9/FA.
; The crt buffer isn't circular, instead the host fills the buffer and the tx
; process drains it, then we can fill it again. Nothing uses 23FA/FB/FC,
; so perhaps that allows for a little bit of overrun.
CMPRS15:			; depth=1
	dcx	h
	dcx	h
	mvi	m,0FBH		; compression token
	inx	h
	mvi	a,' '
	cmp	b
	jz	cmprs16		; use space compression sequence
	mov	m,c		; repeat count
	inx	h
	mov	m,b		; char to repeat
	inx	h
	ret

; emit space compression sequence
cmprs16:
	mvi	a,060H
	add	c
	mov	m,a		; <FB><count+0x60>
	inx	h
	ret

;------------------------------------------------------------------------------
; Time-limited poll for CBS or OBS (about L*50 usec).
; (the loop period is 88 cycles; at 1.78Mhz, this is about 50 uS)
; If CBS or OBS are set, return with z=true (oddly, addr isn't checked)
; otherwise return with z=false.
TIMED_POLL:			; depth=2 (call here, plus a push)
	push	h
	jmp	tpoll2

; Like TIMED_POLL, but limited to 150 usec (three attempts)
POLL_150:			; depth=2 (call here, plus a push)
	push	h
	mvi	l,3		; loop counter
tpoll1:
	in	IN_2200_STATUS
	rrc
	jc	tpoll3		; jump if OBS
	rrc
	jc	tpoll3		; jump if CBS
tpoll2:
	rlc
	rlc
	xri	010H		; flip board select
	ani	0F8H		; 1111 1000
	cmp	d
	jnz	tpoll4		; done if addr changed, !board_sel, or !cpb
	dcr	l
	jnz	tpoll1
	ori	001H		; clear z flag
	pop	h
	ret

tpoll3:
	xra	a		; set z flag
tpoll4:
	pop	h
	ret

;------------------------------------------------------------------------------
; this is called when CBS(nn) occurs at addr 4/5/7
;
; quoting 2236MXE_Documentation.8-83.pdf (pdf page 16)
;    In addition to OBS, with data to be displayed, this address also supports
;    CBS's at this address. When the controller receives a CBS(xx) at address
;    07h it should place a FBh before it in the buffer. This will cause the
;    terminal to accept this string as a command. The following are the three
;    possible commands to the terminal.
;        <FB><count><char> -- repeat <char> <count> times, count 0x00 to 0x50
;        <FB><Cx>          -- x=0..9; delay x/6 seconds at the terminal
;        <FB><D0>          -- literal <FB> character
;
; For addr 4 (printer), compression is not allowed, but <FB> must still be
; expanded to <FB><D0>.
;
; It doesn't mention if addr 5 allows it, but I suspect not.
CBS457:				; depth=2 (call here, rst 2)
	in	IN_OBUS_N
	cma			; non-inverted polarity
	cpi	051H
	jc	cbs457b		; jump if <= 0x50
	cpi	0C0H
	rc			; return if < 0xC0
	cpi	0D0H
	rnc			; return if >= 0xD0
	mvi	m,0FBH		; escape code
cbs457a:
	inx	h		; bump past escape code
	mov	m,a		; save char
	inx	h
	ret

; CBS(nn), nn <= 0x50 means repeat the next OBS char <nn> times.
; it isn't expanded locally; just send the <FB><nn><char> bytes.
cbs457b:
	mvi	m,0FBH		; escape code
	inx	h
	mov	m,a
	di
	lda	SHADOW_RBI
	cma
	ora	e		; mark port(s) ready, as we need to get an OBS
	cma
	rst	2		; save A to SHADOW_RBI, send it to !RBI register, enable interrupts
cbs457c:
	in	IN_2200_STATUS
	rrc
	jnc	cbs457c		; wait for OBS (repeated char value)
	mvi	a,0F9H
	cmp	l
	jnc	cbs457d		; jump if L <= 0xF9 (not end of buffer)
	; hit end of buffer; mark port(s) busy
	lda	SHADOW_RBI
	ora	e		; mark port(s) busy
	rst	1		; save A to SHADOW_RBI, then send it to !RBI register
cbs457d:
	in	IN_OBUS_N	; fetch the OBS value
	cma
	cpi	' '
	jnz	cbs457a		; use three byte <FD><count><char> sequence
	mov	a,m
	adi	060H		; use two byte <FD><count+0x60> space sequence
	mov	m,a
	inx	h
	ret

;------------------------------------------------------------------------------
; Interrupt handler. The mxd uses interrupt only for rx activity.
; verify that the byte was received without error (echo '#' if not).
; keys are received in pairs of (prefix, value) bytes. Once a pair is assembled,
; it is pushed into the keyboard buffer.
;
; If a FC/FD/FE/FF byte is received, it is saved as the prefix byte; the next
; byte received is the associated value and the pair are pushed as-is into the
; KB buffer. But if there is no pending prefix byte, the various values are
; interpreted as follows:
;
;    rx error, 00-07, 09-0C, 0E-0F -> FF F8
;    08          -> FF 08            (backspace)
;    0D          -> FF 0D            (return)
;    12          -> (reset: immediate action)
;    13          -> (halt:  immediate action)
;    10-11,14-7F -> FC <10-11,14-7F>
;    80-BC       -> FD <code>        (81=CLEAR,82=RUN,84=CONTINUE,A0=PRINT,A1=LOAD)
;    BD          -> FD BD            (edit key)
;    BE          -> FD 7E            (lowercase FN)
;    BF          -> FD 7F            (uppercase FN)
;    C0-DF       -> FD <00-1F>       (s.f. keys)
;    E0-F7       -> FF <code>	     (E5=line erase)
;    F8/F9/FA/FB -> (flow control handled immediately, not queued)
;    FC/FD/FE/FF -> (save as prefix for next byte)

INT_HANDLER:
	in	IN_UART_RXRDY
	ani	00FH
	jz	INT_EXIT	; done if something other than RX activity
	call	TERM_LOOKUP	; choose one of the ready RX ports
	mov	a,c
	out	OUT_UART_SEL	; select the chosen UART
	mvi	l,017H		; 2n17: rx sequence prefix byte
	mov	d,m		; see if there was a prefix character
	mvi	m,000H		; forget it for next time
	in	IN_UART_STATUS	; retrieve status associated with char
	xri	080H		; invert DSR bit
	ani	0B8H		; 10111000b = leave DSR/FE/OE/PE bits
	in	IN_UART_DATA	; fetch the character which caused the interrupt
	jm	INT_EXIT	; done if !DSR (received a byte but sender isn't ready to receive)
	jz	int_hand1	; jump if no FE/OE/PE errors
	mvi	a,037H		; 00110111b = RTS + error flag reset + rx enable + dtr + tx enable
	out	OUT_UART_CMD
	jmp	int_hand2

	; we received an error-free character from the terminal (in A)
	; D holds a possible prefix character (00 if none)
int_hand1:
	mov	e,a		; E now holds the rx character
	xra	a
	cmp	d
	jnz	PUSH_KEY_PAIR	; jump if there was a prefix character

	; from here out, we have received a character and no prefix byte.
	ora	e		; A holds the rx character again (saves 1 cycle over mov a,e)
	cpi	0F8H
	jnc	RX_GE_F8	; jump if >= 0xF8  (F8..FB are flow control)
	cpi	014H
	jnc	int_hand4	; jump if >= 0x14
	cpi	012H
	jnc	RX_12_OR_13	; jump if 0x12 or 0x13 (reset or halt key)
	cpi	00DH		; carriage return
	jz	int_hand3
	cpi	008H		; backspace
	jz	int_hand3
	cpi	010H
	jnc	int_hand4	; jump if not a control character
	; fall through ...

	; we received some char with error, or a ...
	; control char other than 0x08 or 0x0D, and no prefix byte (fall through case)
int_hand2:
	mvi	e,0F8H		; pretend we received 0xF8 char
int_hand3:
	mvi	d,0FFH		; set the prefix byte to 0xFF
	jmp	PUSH_KEY_PAIR	; <FF><code> pair, <code> is 08/0D/BD/>=E0, or F8(error)

	; we received 0x10, 0x11, or 0x14 <= char < 0xF8, and no prefix byte
int_hand4:
	cpi	080H
	jc	int_hand6	; jump if < 0x80     --> <FC> <code>
	cpi	0E0H
	jnc	int_hand3	; jump if >= 0xE0    --> <FF> <code>
	cpi	0BDH		; EDIT key
	jc	int_hand5	; jump if < 0xBD     --> <FD> <code>
	jz	int_hand3	; jump if = 0xBD     --> <FF> <BD>
	xri	0C0H		; map SF keys (0xC0-0xDF) to (0x00-0x1F), FN (BE-BF) to (7E-7F)
	mov	e,a
int_hand5:
	; save in key buffer as endi/atom pair (<FD><char>)
	mvi	d,0FDH		; prefix 0xFD = 253
	jmp	PUSH_KEY_PAIR

	; we received 0x10, 0x11, or 0x14 <= char < 0x80
	; save in key buffer as transparent pair (<FC><code>)
int_hand6:
	mvi	d,0FCH		; prefix 0xFC = 252
	; fall through ...

;------------------------------------------------------------------------------
; save a keystroke to the KEYBOARD buffer, which is 32 entries of 2 bytes each.
; a key event is encoded as a pair of bytes with D=prefix byte, E=key.
PUSH_KEY_PAIR:
	lxi	h,KB_BUF_READY
	mov	a,m
	ora	c		; OR in current port bit
	mov	m,a
	mov	h,b
	mvi	l,063H		; 2n63: line_req flag byte
	mov	a,m
	ani	002H		; test KEYIN_LINE_REQ flag
	cnz	SET_KB_RDY_STATUS	; if command is KEYIN_LINE_REQ, set terminal status response indicating a key is ready
	lda	MVP_MODE
	ana	a
	jnz	pkp1		; jump if mvp mode
	mov	a,c
	rrc
	jnc	INT_EXIT	; bail if not term 1
	lda	SHADOW_RBI
	ani	0FEH		; addr 1 (keyboard) is ready
	rst	1		; save A to SHADOW_RBI, then send it to !RBI register
pkp1:
	mvi	l,018H		; 2n18: count of entries in rx kb buffer
	mov	h,b
	mov	a,m		; get count of saved keystrokes
	cpi	020H		; >=32 entries?
	jnc	INT_EXIT	; bail if buffer is full
	inr	m		; increment keystroke count
	inx	h		; 2n19
	inr	m
	inr	m		; bump by 2 (two bytes per rx char)
	mov	a,m
	cpi	05BH		; end of kb buffer? (2n5B is just past the end)
	jc	pkp2		; not wrapping
	mvi	a,01BH		; yes; wrap back to start
	mov	m,a
pkp2:
	mov	l,a
	mov	m,d		; save prefix byte
	inx	h
	mov	m,e		; save char
	; fall through ...

INT_EXIT:
	lda	CUR_UART
	out	OUT_UART_SEL	; restore default UART select
	pop	psw		; clean up (see INTERRUPT push sequence)
	pop	b
	pop	d
	pop	h
	ei
	ret

; we received 0x12 (RESET) or 0x13 (HALT) from the terminal
RX_12_OR_13:
	lxi	h,HALT_MAP
	jnz	RX_RST_HALT	; jump if char is 0x13  (HALT)

	; reset has been pressed
	lda	MVP_MODE
	ana	a
	jnz	RX_RST		; jump if MVP mode

	; we are in VP mode
	mov	a,c
	rrc
	jnc	INT_EXIT	; ignore it if not term 1
	out	OUT_PRIME	; strobe bus reset
	jmp	POWER_ON_INIT

	; reset key pressed in MVP mode
RX_RST:
	; in WANGTERM, clear the bit corresponding to the current term
	mov	a,c
	cma
	lxi	h,WANGTERM
	ana	m
	mov	m,a

	lxi	h,RESET_MAP
	; fall through ...

; HL=RESET_MAP if char was 0x12 or HL=HALT_MAP if it was 0x13
RX_RST_HALT:
	mov	a,m
	ora	c
	mov	m,a		; mark which term hit RESET or HALT
	call	ADDR2_READY	; indicate status has changed
	call	RESET_KB_BUF
	lda	MVP_MODE
	ana	a
	jnz	INT_EXIT	; done if MVP mode
	mov	a,c
	rrc
	jnc	INT_EXIT	; done if not port 1
	lda	SHADOW_RBI
	ori	001H		; addr 1 (KB) is not ready
	rst	1		; save A to SHADOW_RBI, then send it to !RBI register
	out	OUT_HALT_STEP	; strobe real HALT line on bus as we are in VP mode
	jmp	INT_EXIT

;------------------------------------------------------------------------------
; we've received a character >= 0xF8 from the terminal, with no prefix byte
RX_GE_F8:
	cpi	0FCH
	jnc	rxgf8d		; jump if >= 0xFC (not a flow control char)
	; 0xF8-0xFB are flow control characters from the terminal to the MXD
	;    F8: CRT go
	;    F9: PRINT go
	;    FA: CRT stop
	;    FB: PRINT stop
	; these don't go into the rx queue and immediately affect the flow
	; control state.
	lxi	h,PRT_STOPPED	; printer flow control status
	rrc			; bit 0: evens are crt-related, odds are prt-related
	jc	rxgf8a		; jump if printer flow control
	lxi	h,CRT_STOPPED	; crt flow control status
rxgf8a:	rrc			; bit 1: 0 is go, 1 is stop
	mov	a,c		; current term bitmap
	jc	rxgf8b		; jump if "stop" flow control byte
	; we have a "go" flow control byte
	cma
	ana	m		; clear bit for current term
	jmp	rxgf8c

	; we have a "stop" flow control byte
rxgf8b:	ora	m		; set bit for current term
rxgf8c:	mov	m,a		; save updated *_STOPPED byte
	cma			; now 0=stopped, 1=not stopped
	dcx	h		; (PRT/CRT)_BUF_ACTIVE
	ana	m		; clear bit if the buffer is empty
	dcx	h		; (PRT/CRT)_CAN_TX
	mov	m,a		; 0=flow controlled or buffer empty

	; remember that we've seen flow control, meaning the terminal
	; can handle compression and such. this provision seems half-
	; baked, though, as there are so many unconditional dependencies
	lda	WANGTERM
	ora	c		; set bit corresponding to current term
	sta	WANGTERM	; 1=(this term && !flow_controlled && !empty)
	jmp	INT_EXIT

	; received character >= 0xFC from the terminal without prior prefix byte.
	; only FC/FD/FE are documented. but if FF is received, save that too.
	; save the rx byte as the prefix to the next byte.
rxgf8d:
	mov	h,b
	mvi	l,017H
	mov	m,a		; [2n17] = prefix character of some sequence
	jmp	INT_EXIT

;------------------------------------------------------------------------------
; subroutine: if kb rx buffer is empty, set RBI to not ready
CHK_KB_BUF:			; depth=1
	mvi	l,018H		; 2n18: count of entries in rx kb buffer
	di
	mov	a,m
	ana	a
	rnz			; return if buffer is non-empty
	; fall through ...

; subroutine: mark kb rx buffer as empty, and set RBI for addr 1 to not ready
; return with A=0x00 and z flag set (not sure anyone uses this)
MARK_KB_BUF_MT:			; depth=1
	mov	a,c
	cma
	lxi	h,KB_BUF_READY
	ana	m		; clear kb buf ready status for this term
	mov	m,a
	mov	a,c
	rrc
	jnc	mkbmt1		; jump if not term 1
	lda	SHADOW_RBI
	ora	c		; mark not ready
	rst	1		; save A to SHADOW_RBI, then send it to !RBI register
mkbmt1:
	xra	a
	mov	h,b
	ret

;------------------------------------------------------------------------------
; fetch the next entry from the KB rx buffer,
; returning with A=entry prefix byte, E=entry char value.
; if the kb buffer is empty, A=0xFF, E=0x00.
NEXT_KB_PAIR:			; depth=2 (call here, call CHK_KB_BUF)
	mov	h,b		; save 20/24/28/2C page info
	mvi	l,016H		; 2n16
	mov	a,m
	ana	a
	jnz	nkbp2
	call	CHK_KB_BUF	; check if kb buffer is empty
	jz	gkbe4		; jump if it is empty
	dcr	m		; hl=2n18; decrement entry count
	cz	MARK_KB_BUF_MT	; set addr 1 status if newly empty
	mvi	l,01AH		; 2n1A: get pointer of rx buffer
	inr	m
	inr	m		; bump by 2 (two bytes per rx char)
	mov	a,m
	cpi	05BH		; end of kb buffer? (2n5B is just past the end)
	jc	nkbp1		; not wrapping
	mvi	a,01BH		; wrap to start of buffer
	mov	m,a
nkbp1:
	mov	l,a
	mov	a,m		; A=prefix byte
	inx	h
	mov	e,m		; E=char value
	ei
	ret

nkbp2:
	mvi	m,000H
	mov	e,a
	call	CHK_KB_BUF
	ei
	mvi	a,0FCH
	ret

;------------------------------------------------------------------------------
; mark the keyboard buffer as being empty and reenable interrupts
MARK_KB_MT_EI:			; depth=2
	call	MARK_KB_BUF_MT
gkbe4:	; return status indicating empty buffer
	ei
	mvi	a,0FFH
	mvi	e,000H
	ret

;------------------------------------------------------------------------------
; got some kind of unexpected byte or a byte with parity error
; emit '#' to indicate the bad character
BAD_BYTE:
	mvi	a,'#'
	call	A_TO_LINE_REQ
	; fall through ...

; continuation of ERROR_LINE_REQ processing
ERROR_LINE_REQ2:
	; [2n5C] no further continuations
	mov	h,b
	mvi	l,05CH
	mvi	m,000H

	; de = [2n69/6A] (cursor location)
	mvi	l,069H
	mov	d,m
	inx	h
	mov	e,m

	; de=max(cursor location, last byte used), hl=min(...)
	mvi	l,067H
	call	HLPTR_MAXMIN_DE2	; hl <- [2n67/68]; conditionally swap hl/de
	xchg

	; [2n67/68] (last byte used) = max(cursor location, last byte used)
	mov	h,b
	mvi	l,067H		; 2n67
	mov	m,d
	inx	h
	mov	m,e

	mvi	a,007H		; bell
	rst	5		; save bell in the terminal output buffer
	call	RESET_KB_BUF	; ignore any accumulated keys
	; fall through ...

;------------------------------------------------------------------------------
; After processing a key event, the various handlers (mostly) dump back here
; to continue the rx polling process.
RX_POLL1:
	lhld	LINE_REQ_BLOCK
	mov	b,h
	mov	c,m		; one-hot terminal encoding
	call	TEST_CRT_BUF_ACTIVE
rxp2:
	ei
	; fall through ...

RXP3:
	lda	PENDING_TERMS
	ani	00FH
	jz	MAIN_EVENT
	; fall through ...

; pick a terminal with rx chars, grab the next char, and act on it
; the calling routine set the bitmask in A to only allow terms which
; are processing a KEYIN or LINE_REQ command.
RX_POLL:
	; pick a terminal to work on
	call	TERM_LOOKUP	; c=chosen port bitmap, b=h=page0 ptr, a=remaining ports
	sta	PENDING_TERMS	; these uarts have rx data to be unloaded
	mov	a,b
	sta	LINE_REQ_PAGE	; change page of LINE_REQ_BLOCK (20/24/28/2C)

	lxi	h,03FEH		; crt buffer pointer offset
	mov	a,h
	add	b
	mov	h,a		; 0x23FE+n*0x400 = crt put pointer

	; the crt buffer is not circular. downstream logic may copy blocks of bytes
	; from the line_req buffer to the crt buffer, and no end-of-buffer condition
	; is tested. instead, we ensure that there is at least (250-35), or 215, bytes
	; free in the buffer. note we are testing the position of the put ptr, and not
	; (put-get). if the put ptr is too far along, just give up for now and eventually
	; the crt buffer will be drained to the uart tx buffer and the put ptr will get
	; reset to zero.
	mov	a,m		; fetch put pointer
	cpi	023H		; don't process kb events if crt output buffer is too deep
	jnc	RXP3		; jump if >= 35

	; set global underline flag byte if current line_req asked for it
	mov	h,b
	mvi	l,063H
	mov	a,m		; a = [2n63]
	ani	080H		; isolate underline flag
	sta	ULINE_FLAG

	; dispatch to continuation routine if current command isn't complete.
	; any time the continuation byte is set, [KB_BUF_READY] is also set,
	; which ensures this will get called in the RX_POLL loop.
	mvi	l,05CH
	mov	a,m		; a = [2n5C]: continuation byte
	ana	a
	jnz	IDXJMPA

	call	NEXT_KB_PAIR	; a=prefix byte, e=key code
	mov	d,a		; d=prefix byte

	mov	h,b
	mvi	l,063H		; hl=2n63 (flag byte)

	inr	a
	jz	KB_TYPE_FF	; jump if 0xFF (buffer empty or special)
	inr	a
	jz	KB_TYPE_FE	; jump if 0xFE = dead key
	inr	a
	jz	KB_TYPE_FD	; jump if 0xFD = EndI(<=x7f)/Atom(>=x80)
	; must be type FC, which is "transparent" class (nothing special)
	mov	a,e
	cpi	00DH
	jz	L0D43		; jump if carriage return
	call	A_TO_LINE_REQ
rxp4:
	cz	BEEP		; beep if there was no room
	jmp	RX_POLL1

;------------------------------------------------------------------------------
; received a dead key: echo it, then backspace
KB_TYPE_FE:
	mov	a,e
	cpi	008H
	jz	kbtfe2		; jump if backspace
	call	A_TO_LINE_REQ
	jz	rxp4		; z means no room: beep and quit
kbtfe1:
	lxi	d,0FFFFH	; -1: move cursor left 1 place
	jmp	CUR_LEFT_DE

kbtfe2:
	; <FE><08> sequence
	mvi	a,' '
	call	A_TO_LINE_REQ
	jz	DO_BS		; z means no room
	jmp	kbtfe1

;------------------------------------------------------------------------------
; <FF><code> pair pulled from kb buffer
; inspecting the table in the comment before INT_HANDLER,
; the only <code> bytes we should expect are 08, 0D, E0-F7.
KB_TYPE_FF:
	mov	a,e
	cpi	010H
	jc	KB_FF_CTRL	; jump if a control code
	cpi	0BDH
	jz	KB_FF_EDIT	; jump if EDIT key
	cpi	0E6H		; key for statement number
	jc	KB_FF_LT_E6	; jump if < 0xE6
	jnz	BAD_BYTE	; jump if not 0xE6
	mvi	a,0FEH		; map to key pair <FF><FE> (TBD what does this mean?)
	jmp	L0D43		; report <FF><FE> (it was <E6>)

;------------------------------------------------------------------------------
; stuff A into line request buffer, updating the (cursor location) and
; (last byte used) pointers, as well as the dirty span.
; return with Z flag set if no room for it, or if CR is passed in
A_TO_LINE_REQ:			; depth=3 (call here, call A_TO_CRT_BUF)
	cpi	00DH
	rz			; ignore carriage return
	push	psw		; save new character

	; de =[2n69/6A]
	mov	h,b
	mvi	l,06AH
	mov	e,m
	dcx	h
	mov	d,m

	xchg
	mov	m,a		; store character to location pointed to by [2n69/6A]
	xchg

	; jump to atlr2 if [2n67/68](last used byte) != [2n69/6A](cursor location)
	dcx	h
	mov	a,m
	cmp	e
	jnz	atlr2
	dcx	h
	mov	a,m
	cmp	d
	jnz	atlr2

	; if [2n65/66](end of buffer) == [2n69/6A](cursor location), ignore request
	dcx	h
	mov	a,m
	cmp	e
	jnz	atlr1
	dcx	h
	mov	a,m
	cmp	d
	jz	POPRET		; give up

	; the cursor is at the last byte used, but we aren't at the end of buffer
atlr1:
	; increment [2n67/68](last byte used)
	inx	d
	mvi	l,067H
	mov	m,d
	inx	h
	mov	m,e
	dcx	d		; revert de
	; fall through ...

atlr2:
	; increment [2n69/6A](cursor location)
	inx	d
	mvi	l,069H
	mov	m,d
	inx	h
	mov	m,e

	; echo the character to the crt buffer
	pop	psw		; recall original character value
	call	A_TO_CRT_BUF	; echo the character

	; [2n64]++ (bump cursor column)
	mvi	l,064H
	inr	m
	; fall through ...

;------------------------------------------------------------------------------
; wrap the line_req cursor column if it is 80
; hl=2n64 on entry
WRAP_CUR_COL:
	mov	a,m		; fetch current line_req cursor column
	cpi	80
	rc			; return if column didn't wrap
	mvi	m,000H		; [2n64] (cursor column) = 0
	mvi	a,00AH		; issue linefeed
	rst	5		; push it to the terminal output buffer
	dcr	a		; clear Z flag
	ret

;------------------------------------------------------------------------------
; <FF><code> pair pulled from kb buffer, and <code> < 0xE6
; erase the line_req field, but if the 0x20 bit of the flag byte
; was not set, clear the edit mode state.
KB_FF_LT_E6:
	mvi	e,06BH		; start of line_req buffer
	mov	d,b
	call	CURS_TO_DE	; move cursor to start of buffer
	call	CLR_CUR_TO_EOB	; clear buffer to the right of the cursor
	mov	h,b
	mvi	l,063H		; 2n63: line_req flag byte
	mov	a,m
	ani	020H		; linput statement active
	jnz	RX_POLL1

	; set the edit mode bit, but the next section is going to invert it
	; so really this is setting it up to be cleared
	mvi	a,004H
	ora	m
	mov	m,a
	; fall through ...

; flip the line_req edit mode bit; change cursor blink mode to correspond
; arrived here due to <FF><EDIT> key pair, or fall-through
KB_FF_EDIT:
	mvi	a,0FBH		; compression token
	rst	5		; push it to the terminal output buffer
	mov	a,m
	xri	004H		; flip the edit mode bit
	mov	m,a
	ani	004H		; isolate the edit mode bit
	adi	0F8H		; 0xF8 (no cursor blink) or 0xFC (cursor blink)
	jmp	DOBS3		; send A to crt buffer

;------------------------------------------------------------------------------
; KB input was a control character of some kind (<0x10)
KB_FF_CTRL:
	cpi	00DH		; carriage return
	jz	L0D43
	cpi	008H		; back space
	jnz	RXP3		; ignore it
	; fall through ...

;------------------------------------------------------------------------------
; process a backspace request
DO_BS:
	; de = [2n69/6A] (cursor location)
	mov	h,b
	mvi	l,069H
	mov	d,m
	inx	h
	mov	e,m

	; branch if cursor location==2n6B (start of buffer)
	mvi	a,06BH
	sub	e
	jnz	dobs1
	ora	h
	cmp	d
	jz	RXP3		; jump if at start of buffer

dobs1:
	; jump if [2n67/68] (last byte used) != de
	mvi	l,067H
	mov	a,m
	cmp	d
	jnz	dobs2
	inx	h
	mov	a,m
	cmp	e
	jnz	dobs2

	; [2n67/68](last byte used) = (cursor location - 1)
	dcx	d
	mov	m,e
	dcx	h
	mov	m,d
	inx	d
dobs2:
	dcx	d		; move cursor pointer left one
	xchg
	mvi	m,' '
	call	CURS_TO_HL
	mvi	a,' '
	call	A_TO_CRT_BUF
	mvi	a,008H		; backspace
DOBS3:
	rst	5		; save A in the terminal output buffer
	jmp	RX_POLL1

;------------------------------------------------------------------------------
; received s.f. 8 in EDIT mode (erase from cursor to end of buffer)
EDIT_SF8:
	call	CLR_CUR_TO_EOB
	jmp	RX_POLL1

; clear the line_req buffer from the cursor to the end of buffer
CLR_CUR_TO_EOB:
	; de = [2n69/6A] (cursor location)
	mov	h,b
	mvi	l,069H
	mov	d,m
	inx	h
	mov	e,m
	push	d		; why? de isn't changed between here and the pop

	; hl = [2n67/68] (last byte used), then update
	; [2n67/68] (last byte used) = de = (cursor location)
	mvi	l,067H
	mov	a,m
	mov	m,d
	inx	h
	mov	c,m
	mov	m,e
	mov	h,a
	mov	l,c

	pop	d		; == cursor location
	call	HL_SUB_DE	; hl = distance from cursor to last byte used
	; fall through ...

;------------------------------------------------------------------------------
; hl = count of spaces (possibly underlined) to emit to the terminal
; the cursor will be moved back to the starting point afterwards
EMIT_HL_SPACES:
	push	h
	xchg
	mov	h,b
	mvi	l,064H		; 2n64: line_req cursor column
	mov	a,m
	cma
	adi	81
	mov	c,a		; c=80-cursor_column
	xchg			; hl=distance from cursor to last byte used

ehls1:	; draw the minimum of current cursor location to column 80, or to residual hl count
	xra	a
	mov	d,a
	mov	e,c		; de=distance from cursor to column 80
	cmp	h
	jnz	ehls2
	ora	l
	cmp	c		; (cursor to last used byte) - (cursor to col 80)
	jnc	ehls2		; jump if (cursor to last byte used) >= (cursor to col 80)
	mov	e,l		; de is now the smaller of the two
ehls2:
	mov	a,c		; distance of cursor to col 80
	sub	e
	mov	c,a
	call	HL_SUB_DE	; hl=(current clear point to last byte used)
	mvi	a,2
	cmp	e
	jnc	ehls4		; if run was < 3 chars, don't use compression
	mvi	a,0FBH		; compression token
	rst	5		; push compression token to the terminal output buffer
	lda	ULINE_FLAG
	ana	a
	jnz	ehls3		; if underline mode is in effect, use three-byte compression
	mvi	a,060H		; we can use two byte compression:
	add	e		; FB nn, where nn >= 0x60 means (nn-0x60) spaces
	rst	5		; push A to the terminal output buffer
	jmp	ehls6

	; use three byte compression for underlined spaces: FB nn <char>
ehls3:
	mov	a,e
	rst	5		; push the repetition count to the terminal output buffer
	mvi	a,0A0H		; underlined space
	rst	5		; push A to the terminal output buffer
	jmp	ehls6

	; emit E spaces (possibly underlined) without compression
ehls4:
	xra	a
	cmp	e
	jz	ehls6		; nothing to erase
ehls5:
	lda	ULINE_FLAG
	ori	' '		; possibly underline the space
	rst	5		; push it to the terminal output buffer
	dcr	e
	jnz	ehls5
	; fall through ...

	; all characters to the right of the cursor have been cleared on the terminal,
	; but the buffer might extend beyond the first row. clear chars on subsequent
	; rows if necessary.
ehls6:
	xra	a
	cmp	c
	jnz	ehls7
	mvi	a,00AH		; linefeed
	rst	5		; push it to the terminal output buffer
	mvi	c,80		; if we have do do another line, we will clear 80 columns
ehls7:
	mov	a,l
	ora	h
	jnz	ehls1		; still work left

	mvi	a,80
	sub	c
	mov	h,b
	mvi	l,064H		; 2n64: line_req cursor column
	mov	m,a

	pop	d		; distance from original cursor to last byte used
	mov	a,d		; \
	cma			;  \
	mov	d,a		;   \
	mov	a,e		;    > two's complement of de
	cma			;   /
	mov	e,a		;  /
	inx	d		; /

	jmp	CTDE1		; move the cursor back to where the clear started

;------------------------------------------------------------------------------
; received s.f. 15 (recall) in EDIT mode
EDIT_SF15:
	; de = [2n69/6A] (cursor location)
	mvi	l,069H
	mov	d,m
	inx	h
	mov	e,m

	; [2n67/68] (last used byte) =?= [2n68/6A] (cursor location)
	mvi	l,067H
	mov	a,m
	xra	d
	jnz	RX_POLL1
	inx	h
	mov	a,m
	xra	e
	jnz	RX_POLL1	; jump to RX_POLL1 if (last byte used) != (cursor location)
			; perhaps don't attempt recall if pointers don't match
			; for instance, recall when editing lines has no effect unless the cursor is before the first non-space character of the buffer
			; ... but even that isn't right, in that we can edit the line number itself, so the buffer really starts before the first non-space char after the line number.
	mvi	a,0FFH
	; fall through ...
	; but it is weird because d=2n, and isn't a prefix byte

;------------------------------------------------------------------------------
; do something with the rx pair; d=prefix byte and a=keycode
; it is called when a key is processed that the line_req processing can't handle itself:
;     KB_FF_CTRL   jumps here with <FF><0D>
;     RX_POLL      jumps here with <FC><0D>
;     INJECT_CR    jumps here with <??><0D>
;     KB_TYPE_FF   jumps here with <FF><FE> (in response to <FF><E6>=STMT# key)
;     KB_TYPE_ENDI jumps here with <FD><nn>, where 0x60 <= nn <= 0x7F (s.f. key)
;     KB_TYPE_ENDI jumps here with <FD><nn>, where 0x20 <= nn <= 0x3F (s.f. key)
;     KB_TYPE_ENDI jumps here with <FD><nn>, where 0x00 <= nn <= 0x1F (s.f. key) and line_req edit_mode bit is set
L0D43:
	; QUERY_LINE_REQUEST uses this state to know ...
	; ... when to return FFh (see that command's description)
	mov	h,b
	mvi	l,061H
	mov	m,d		; [2n61] = d = key prefix byte
	inx	h
	mov	m,a		; [2n62] = a = keypress value

	; de = [2n69/6A] (cursor location)
	mvi	l,069H
	mov	d,m
	inx	h
	mov	e,m

	; [2n5D/5E] (redraw start) = [2n69/6A] (cursor location)
	mvi	l,05DH
	mov	m,d
	inx	h
	mov	m,e

	; [2n5F/60] (redraw end) = [2n69/6A] (cursor location)
	inx	h
	mov	m,d
	inx	h
	mov	m,e

	call	SET_KB_RDY_STATUS
	lxi	h,LREQ_ACCEPT	; map, 1=line_req active
	mov	a,c
	cma
	ana	m		; clear bit for the current terminal
	mov	m,a
	jmp	rxp2

;------------------------------------------------------------------------------
; <FD><code lt 0x80> key sequence received.
; d=prefix byte, e=key code
; it is described as a special function key received, but the code below is far
; more complicated than that.
KB_TYPE_ENDI:
	cpi	060H
	jnc	L0D43		; jump if A >= 0x60
	cpi	040H
	jnc	kbte1		; jump if A >= 0x40 (and < 0x60)
	cpi	020H
	jnc	L0D43		; jump if A >= 0x20 (and < 0x40)
	mov	a,m		; a = [2n63]: line_req flag byte
	ani	004H		; test EDIT mode flag
	mov	a,e		; save key code
	jz	L0D43		; jump if not edit mode

; either <FD><cc> where 0x40 <= cc < 0x60
;     or <FD><cc> where cc < 0x20 in edit mode
kbte1:	; 0x01-0x1F or 0x20-0x3F
	ani	00FH
	sui	004H
	jc	RXP3		; jump if s.f. 0-3; ignore it
	sui	007H
	jc	EDIT_SF4_10	; jump if s.f. 4-10
	jz	EDIT_SF11	; jump if s.f. 11  (move cursor right 5)
	dcr	a
	jz	EDIT_SF12	; jump if s.f. 12  (move cursor right 1)
	dcr	a
	lxi	d,0FFFFH	; -1
	jz	CUR_LEFT_DE	; jump if s.f. 13  (move cursor left 1)
	dcr	a
	jnz	EDIT_SF15	; jump if s.f. 15  (recall)
	; fall through ...

; received s.f. 14 (move cursor left 5) in EDIT mode
; EDIT_SF14:
	lxi	d,0FFFBH	; -5
	; fall through ...

; adjust the cursor DE places (de is signed)
CUR_LEFT_DE:
	; hl=[2n69/6A]
	mov	h,b
	mvi	l,069H
	mov	a,m
	inx	h
	mov	l,m
	mov	h,a

	dad	d		; move left or 1 or 5 places
	xchg
	mov	h,b
	mvi	l,06BH		; 2n6B: start of buffer
	call	MAX_HL_MIN_DE	; don't let cursor move to the left of the beginning
	call	CURS_TO_HL
	jmp	RX_POLL1

; received s.f. key 4 to 10 in EDIT mode
EDIT_SF4_10:
	inr	a
	jz	EDIT_SF10	; insert
	inr	a
	jz	EDIT_SF9	; delete
	inr	a
	jz	EDIT_SF8	; erase
	inr	a
	jz	EDIT_SF7	; start of line
	inr	a
	jz	EDIT_SF6	; up a line
	inr	a
	jz	EDIT_SF5	; down a line
	; fall through ...

; received s.f. key 4 (end) in EDIT mode
; start at the end of buffer, then scan back until either the start of buffer
; is found or a non-space character is found
; EDIT_SF4:
	; de = [2n67/68] (last byte used)
	mvi	l,067H
	mov	d,m
	inx	h
	mov	e,m

	mvi	c,' '
sf4a:
	; done if (last byte used) is at the start of buffer
	mvi	a,06BH		; 2n6B is start of line req buffer
	sub	e
	jnz	sf4b
	ora	d
	cmp	b
	jz	CUR_TO_DE_DONE	; jump if at start of buffer
sf4b:
	dcx	d		; back up one char
	ldax	d		; fetch it from line_req buffer
	cmp	c
	jz	sf4a		; jump if space
	inx	d		; go back one
	jmp	CUR_TO_DE_DONE

; received s.f. 7 (begin) in EDIT mode
EDIT_SF7:
	call	CURS_TO_SOL
	jmp	RX_POLL1

; received s.f. 11 (right cursor 5) in EDIT mode
EDIT_SF11:
	lxi	d,5
	jmp	CUR_RIGHT_DE

; received s.f. 12 (right cursor 1) in EDIT mode
EDIT_SF12:
	lxi	d,1
	; fall through ...

; move the cursor right DE places
CUR_RIGHT_DE:
	; hl = [2n69/6A] (cursor location)
	mov	h,b
	mvi	l,069H
	mov	a,m
	inx	h
	mov	l,m
	mov	h,a

	dad	d		; adjust right 1 or 5
	xchg
	mvi	l,067H
	call	HLPTR_MAXMIN_DE	; hl = [2n67/68](last byte used); hl=max, de=min
	; fall through ...

; move cursor to the line_req buffer offset pointed at by de,
; then resume timeslicing
CUR_TO_DE_DONE:
	call	CURS_TO_DE
	jmp	RX_POLL1

;------------------------------------------------------------------------------
; received s.f. 5 (down) in EDIT mode
EDIT_SF5:
	; hl = [2n69/6A] (cursor location)
	mov	h,b
	mvi	l,069H
	mov	d,m
	inx	h
	mov	e,m

	lxi	h,80
	dad	d		; hl += 80

	xchg
	mvi	l,067H		; 2n67
	call	HLPTR_MAXMIN_DE	; hl = [2n67/68](last byte used); hl=max, de=min
	cnc	CURS_TO_DE	; call if we swapped above (don't move if we tried to go past end of line)
	jmp	RX_POLL1

;------------------------------------------------------------------------------
; received s.f. 6 (up) in EDIT mode
EDIT_SF6:
	; de = [2n69/6A] (cursor location)
	mov	h,b
	mvi	l,069H
	mov	d,m
	inx	h
	mov	e,m

	lxi	h,0FFB0H
	dad	d		; hl = (cursor location) - 80

	mov	d,b
	mvi	e,06BH
	call	MAX_HL_MIN_DE	; de = min(start of buff, cur_loc-80)
	cnc	CURS_TO_HL	; if we didn't undershoot, move the cursor
	jmp	RX_POLL1

;------------------------------------------------------------------------------
; received s.f. 10 (insert) in EDIT mode
; this shifts the contents of the line_req buffer, but other code does the
; redraw to the terminal (also true of most other commands)
EDIT_SF10:
	mov	h,b
	mvi	a,061H		; continuation byte
	call	SAV_CONT

	; de = [2n67/68] (last byte used)
	mov	h,b
	mvi	l,067H
	mov	d,m
	inx	h
	mov	e,m

	; if de (last byte used) != [2n65/66] (end of buffer), increment (last byte used)
	; that is, inserting will increment the last byte used pointer
	; funny that the jump instructions jump to the rewrite of [2n67/68] instead of
	; just skipping that part of it.
	mvi	l,066H
	mov	a,m
	cmp	e
	jnz	sf10a
	dcx	h
	mov	a,m
	cmp	d
	jz	sf10b
sf10a:
	inx	d		; increment last byte used pointer
sf10b:
	; [2n67/68] (line_req last byte used) = de
	mvi	l,067H
	mov	m,d
	inx	h
	mov	m,e

	; [2n5F/60] = de = copy of last byte used
	mvi	l,05FH
	mov	m,d
	inx	h
	mov	m,e
	push	d

	; de = [2n69/6A] (cursor location)
	mvi	l,069H
	mov	d,m
	inx	h
	mov	e,m

	; [2n5D/5E] = de = [2n69/6A] (cursor location)
	mvi	l,05DH
	mov	m,d
	inx	h
	mov	m,e

	xchg		; hl = cursor location
	pop	d	; de = last byte used
	mvi	c,' '	; we want to erase at the insertion point
	jmp	sf10d

sf10c:
	mov	a,m	; read byte we are about to overwrite
	mov	m,c	; write the byte that was to the left
	mov	c,a	; prepare for next loop
	inx	h
sf10d:
	; loop until de=hl
	mov	a,l
	cmp	e
	jnz	sf10c
	mov	a,h
	cmp	d
	jnz	sf10c
	jmp	RX_POLL1

;------------------------------------------------------------------------------
; received s.f. 9 (delete) in EDIT mode
EDIT_SF9:
	; de = [2n69/6A] (cursor location)
	mov	h,b
	mvi	l,069H
	mov	d,m
	inx	h
	mov	e,m

	; [2n5D/5E] = de = cursor location
	mvi	l,05DH
	mov	m,d
	inx	h
	mov	m,e

	; [2n67/68] (last byte used) =?= (cursor location)
	mvi	l,068H
	mov	a,m
	cmp	e
	dcx	h
	jnz	sf9a
	mov	a,m
	cmp	d
	jz	RX_POLL1	; jump if (cursor location) = (last byte used)
sf9a:
	push	d	; cursor location

	; de = [2n67/68] (last byte used)
	mov	d,m
	inx	h
	mov	e,m

	; [2n5F/60] = de = last byte used
	mvi	l,05FH
	mov	m,d
	inx	h
	mov	m,e

	; decrement [2n67/68] (last byte used)
	dcx	d
	mvi	l,067H
	mov	m,d
	inx	h
	mov	m,e

	inx	d	; restore to old last byte used location to en

	mvi	a,061H		; continuation byte (C2 = DO_REDRAW)
	call	SAV_CONT
	xchg			; de=..., hl=old last byte used pointer
	pop	d		; de=cursor location, hl=old last byte used
	mvi	c,' '		; fill last used byte with space
	jmp	sf9c

; move the characters one to the left from last used to cursor location
sf9b:
	dcx	h
	mov	a,m
	mov	m,c
	mov	c,a
sf9c:
	; loop until hl=de
	mov	a,l
	cmp	e
	jnz	sf9b
	mov	a,h
	cmp	d
	jnz	sf9b
	jmp	RX_POLL1

;------------------------------------------------------------------------------
; continuation#1 from DO_REDRAW via alrd6
SET_LINEREQ_STATUS:
	mov	h,b
	call	SET_KB_RDY_STATUS	; indicate line_req complete for status reponse
	call	SEND_CR_LF
	; fall through ...

; continuation#3 of EDIT_SF9/10 + continuation#5 of END_OF_LINE_REQ
; if the keyboard buffer is empty, set flags indicating it
CHECK_KB_MT_FLAG:
	; [2n5C] no more continuations
	mov	h,b
	mvi	l,05CH
	mvi	m,000H

	mvi	l,018H		; a = [2n18] (count of entries in rx kb buffer)
	mov	a,m
	ana	a
	cz	MARK_KB_MT_EI	; mark the rx buffer empty if it is and reenable interrupts
	jmp	RX_POLL1

;------------------------------------------------------------------------------
; continuation #1 of END_OF_LINE_REQ and TERMINATE_LINE_REQ
; (via INDTAB branch table and the 2n5C byte)
; enable the cursor and set it to blink (if edit mode) or nonblink (if not edit)
CURSOR_ON:
	; [2n5C]++ continue at BLANK_SPAN next time through event loop
	mov	h,b
	mvi	l,05CH
	inr	m

	lda	WANGTERM
	ana	c
	jz	RX_POLL1	; done if not a Wang terminal

	; enable cursor code (<05>)
	mvi	a,005H
	rst	5

	; make cursor either not blink (<FB><F8>) or blink (<FB><FC>)
	mvi	a,0FBH		; compression token
	rst	5
	mvi	l,063H
	mov	a,m		; [2n63] = line_req flag byte
	ani	004H		; test edit mode flag
	adi	0F8H		; 0xF8 (no blinking) or 0xFC (blinking)
	rst	5
	jmp	RX_POLL1

;------------------------------------------------------------------------------
; continuation #2 of END_OF_LINE_REQ and TERMINATE_LINE_REQ
; (via INDTAB branch table and the 2n5C byte)
; move cursor to start of redraw span and emit spaces to the end of span
BLANK_SPAN:
	; [2n5C]++ continue to DO_REDRAW at next event loop
	mov	h,b
	mvi	l,05CH
	inr	m

	; de = [2n5D/5E] (redraw start)
	inx	h
	mov	d,m
	inx	h
	mov	e,m

	; hl = [2n69/6A] (cursor location)
	mvi	l,069H
	mov	a,m
	inx	h
	mov	l,m
	mov	h,a

	call	HL_SUB_DE	; hl = (cursor)-(redraw start)
	push	h
	call	CURS_TO_DE	; move to redraw start point
	pop	h
	call	EMIT_HL_SPACES
	jmp	RX_POLL1

;------------------------------------------------------------------------------
; push the characters from (cursor location) to (end of redraw span).
; continuation of EDIT_SF9/SF10
; continuation #3 of END_OF_LINE_REQ
; continuation after alrd6
DO_REDRAW:
	call	ALRD_REDRAW
	jmp	RX_POLL1

; ACCEPT_LINE_REQ_DATA (alrd6) leads here, and DO_REDRAW continuation uses this.
; scan the line_req buffer from the cursor location to the end of the redraw
; span, [2nF5/60]. copy each character to the tx output buffer, taking care
; to map control characters to '.' and to escape literal 0xFB characters.
;
; because the crt buffer is smaller than the size of the redraw span, this
; routine will redraw from the current cursor column up to column 80, which means
; a maximum of 80 characters, which might mean 160 bytes in the worst case
; (the worst case would be a string of nothing but <FB>, and each byte would need
; to be escaped individually). If there is work left, the continuation code will
; remain set to this routine, and the next piece will get drawn.
ALRD_REDRAW:			; depth=3 (call here, push h, call DE_SUB_HL)
	mov	h,b
	mvi	l,064H		; 2n64: line_req cursor column
	mvi	a,80
	sub	m
	mov	c,a		; c=80-cursor_col

	; de = [2n5F/60] (end of redraw span)
	mvi	l,05FH
	mov	d,m
	inx	h
	mov	e,m

	; hl = [2n69/6A] (cursor location)
	mvi	l,069H
	mov	a,m
	inx	h
	mov	l,m
	mov	h,a
	push	h

	call	DE_SUB_HL	; hl = de-hl = (end of redraw span) - (cursor location)
	jnz	alrdd1		; jump if distance >= 256; lots left
	mov	a,c
	cmp	l		; (cols left to 80) - (buffer left to redraw)
	jc	alrdd1		; jump if the latter is larger
	mov	c,l		; c=smaller of the two, which means we are done

	; we are about to complete the transfer to (end of redraw span)
	; [2n5C] bump the continuation byte
	;    DO_REDRAW -> SET_LINEREQ_STATUS (if via alrd6)
	;    DO_REDRAW -> CURSOR_TO_START    (all other users)
	mov	h,b
	mvi	l,05CH
	inr	m

	xra	a
	cmp	c
	jnz	alrdd1
	pop	h		; it is already done; over and out
	ret

alrdd1:
	; de = output buffer put pointer
	lxi	h,03FEH
	mov	a,h
	add	b
	mov	h,a		; hl=23FE+n*0x400
	mov	d,h
	mov	e,m

	; a = this term one-hot code
	mov	h,b
	mvi	l,000H		; 2n00
	mov	a,m

	; [CRT_BUF_ACTIVE] |= 1hot_term
	lxi	h,CRT_BUF_ACTIVE
	di
	ora	m
	mov	m,a

	; [CRT_CAN_TX] = ~[CRT_STOPPED] & [CRT_BUF_ACTIVE]
	inx	h		; CRT_STOPPED
	mov	a,m
	cma
	dcx	h		; CRT_BUF_ACTIVE
	ana	m
	dcx	h		; CRT_CAN_TX
	mov	m,a

	ei

	; regC is the number of bytes we are sending to the crt buffer
	; [2n64] (cursor column) += number of bytes we are sending to crt buffer
	; we know there are enough bytes for it in the crt buffer because
	; RX_POLL wouldn't have called the continuation routine unless there
	; was lots of room (the crt put ptr must be less than 35).
	mov	h,b
	mvi	l,064H
	mov	a,m
	add	c
	mov	m,a

	pop	h		; hl = cursor location
alrdd2:
	lda	ULINE_FLAG
	ora	m		; read next byte from line_req buffer
	inx	h
	cpi	010H		; is it a control character?
	jnc	alrdd3
	mvi	a,'.'
alrdd3:
	stax	d		; save char in output buffer
	inx	d		; bump put pointer
	cpi	0FBH		; are we are sending literal FB?
	jnz	alrdd4		; jump if not
	mvi	a,0D0H		; we need to escape it
	stax	d
	inx	d
alrdd4:
	dcr	c
	jnz	alrdd2

	push	h		; save line_req get pointer
	mov	h,d
	mvi	l,0FEH
	mov	m,e		; save output buffer put pointer
	pop	d

	; [2n69/6A] (cursor location) = de
	mov	h,b
	mvi	l,069H
	mov	m,d
	inx	h
	mov	m,e

	mvi	l,064H		; 2n64: line_req cursor column
	jmp	WRAP_CUR_COL

;------------------------------------------------------------------------------
; continuation#2 of EDIT_SF9/10 or continuation#4 of END_OF_LINE_REQ
; move the cursor to the start of the redraw region (if not REFILL_LINE_REQ)
CURSOR_TO_START:
	; [2n5c] bump continuation
	mov	h,b
	mvi	l,05CH
	inr	m

	; de = [2n5D/5E] (redraw start)
	inx	h
	mov	d,m
	inx	h
	mov	e,m

	; a = [2n63] (line_req flag byte)
	mvi	l,063H
	mov	a,m

	ani	010H		; refill_line_req flag
	push	psw
	xra	m		; invert refill_line_req flag
	mov	m,a
	pop	psw
	cz	CURS_TO_DE	; refill (unlike prefill) doesn't move cursor to the start of buffer
	jmp	RX_POLL1

; continuation#5 of TERMINATE_LINE_REQ
; pretend the <CR> key has been seen
INJECT_CR:
	mvi	a,00DH		; carriage return
	jmp	L0D43

; clear the keyboard buffer for this term and mark the kb status as empty
RESET_KB_BUF:
	di
	mov	h,b
	call	INIT_KB_BUF
	mvi	l,000H
	mov	a,m		; [2n00] = one-hot encoding of current term
	cma
	lxi	h,KB_BUF_READY
	ana	m
	mov	m,a		; clear kb buffer ready bit for the current terminal
	ret

; we have a byte which is >= 0x80.
;  "EndI/Atom - <EndI/Atom><Char> where <EndI/Atom> = HEX(FD)
;   MXD response: Char is treated as a special function key if less than 0x80.
;   The MXD is responsible for deatomizing the atom code codes in the case of
;   INPUT or LINPUT or supplying the proper atom code to KEYIN. The atom code
;   sent from the terminal to the MXD need not be the same as the atom code
;   sent from the MXD to the CPU."
; if it matches an entry in the atom table, expand it into the string literal
; into the keyboard buffer
KB_TYPE_FD:
	mov	a,e
	ora	a
	jp	KB_TYPE_ENDI	; jump if < 0x80
	lxi	h,CLEAR_TOK	; start of atom/string table
kbtfd1:
	cmp	m		; compare to atom byte
	inx	h		; hl now points to start of corresponding string
	jz	kbtfd3		; jump if the byte matched the atom in the table
	jnc	kbtfd1		; keep scanning table
	jmp	BAD_BYTE	; must have gotten an atom byte > than byte in A

kbtfd2:
	push	h
	call	A_TO_LINE_REQ	; stuff A into line request buffer
	cz	BEEP		; signal if there is no room left in buffer
	pop	h
	inx	h		; point to next byte in the atom/string table
	; found a table entry matching the atom byte
kbtfd3:
	mov	a,m
	ana	a
	jp	kbtfd2		; string bytes are all < 0x80
	jmp	RX_POLL1	; must have hit the next atom byte or sentinel

; atom/string table, to allow expanding keyboard atoms to simple text
; note: entries are stored in ascending order
; note: msb is used to distinguish atom byte vs following expansion
CLEAR_TOK:
	db	081H
	db	"CLEAR"

	db	082H
	db	"RUN"

	db	084H
	db	"CONTINUE"

	db	0A0H
	db	"PRINT "

	db	0A1H
	db	"LOAD "

	; end of table sentinel
	db	0FFH
	db	0FFH

	nop
	nop
	nop
	nop
	db	042H	; this is forced to produce a 00 checksum

; vim: ts=8:noet:sw=8
